(* ========================================================================= *)
(*                                                                           *)
(*               Imaging Optical Systems and Cardinal Points                 *)
(*                                                                           *)
(*        (c) Copyright, Umair Siddiquen=, 2012-2015                         *)
(*                       Hardware Verification Group (HVG),                  *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last update: September 10,2015                                            *)
(*                                                                           *)
(* ========================================================================= *)

needs "composed_system.ml";;

let asm_real_prove = REPEAT(POP_ASSUM MP_TAC) THEN CONV_TAC REAL_FIELD;;


let gen_optical_system = define` 
    gen_optical_system (sys:optical_system) do dn ni nt = 
    [([],ni,do);sys;([],nt,dn)]`;;

let GEN_OPTICAL_SYSTEM_MAT = prove(` 
    !sys do dn ni nt A B C D. 
    ((system_composition sys) = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D)) ==> 
     composed_system (gen_optical_system sys do dn ni nt) = 
     cmat2x2 (Cx(A + C*dn)) (Cx (A*do + B + C*do*dn + D*dn)) 
            (Cx C) (Cx (C*do + D))`, 

REPEAT STRIP_TAC THEN
REWRITE_TAC[composed_system;gen_optical_system;
            system_composition;CMAT2X2_LID;free_space_matrix] THEN
ASM_REWRITE_TAC[] THEN SIMP_TAC[ CMAT2X2_MUL;CMAT2X2_EQ] THEN
REWRITE_TAC[GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN 
CONV_TAC REAL_FIELD);;


let  GEN_OPTICAL_SYSTEM_MATRIX = prove(`!sys (cray:gray) do dn ni nt A B C D.
   ((system_composition sys) = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D)) /\
   is_valid_optical_system sys  /\ 
   &0 < ni /\  &0 < nt /\  
   is_valid_genray cray /\  
   is_valid_gray_in_system cray (gen_optical_system sys do dn ni nt)
  ==>
  (let y0,theta0 = fst_single_ray (HD cray)
              and y_n,theta_n = last_single_ray (LAST cray) in
   vector [Cx y_n; Cx theta_n] =
         cmat2x2 (Cx(A + C*dn)) (Cx (A*do + B + C*do*dn + D*dn)) 
            (Cx C) (Cx (C*do + D))  ** vector [Cx y0; Cx theta0])`,

SIMP_TAC[GSYM GEN_OPTICAL_SYSTEM_MAT] THEN
IMP_REWRITE_TAC[COMPOSED_SYSTEM_MATRIX] THEN
SIMP_TAC[gen_optical_system;is_valid_composed_system;ALL;
           is_valid_optical_system;is_valid_gray_in_system;is_valid_free_space]);;

(*----------------------------------------------------------*)
(*------------------Imaging Properties----------------------*)
(*----------------------------------------------------------*)

let object_height = define ` 
    object_height (r:ray) = FST  (fst_single_ray r)`;;

let image_height = define ` 
    image_height (r:ray) = FST  (last_single_ray r)`;;

let lastray_height = define ` 
    lastray_height (r:ray) = FST  (last_single_ray r)`;;

let fstray_height = define ` 
    fsttray_height (r:ray) = FST  (fst_single_ray r)`;;

let object_angle = define ` 
    object_angle (r:ray) = SND  (fst_single_ray r)`;;

let image_angle = define ` 
    image_angle (r:ray) = SND  (last_single_ray r)`;;


let object_height_genray = define ` 
    object_height_genray (r:gray) = FST  (fst_single_ray (HD r))`;;


let image_height_gray = define ` 
    image_height_gray (r:gray) = FST  (last_single_ray (LAST r))`;;


let object_angle_genray = define ` 
    object_angle_genray (r:gray) = SND  (fst_single_ray (HD r))`;;

let image_angle_genray = define ` 
    image_angle_genray (r:gray) = SND  (last_single_ray (LAST r))`;;





let lateral_magnification_gray = define `
     lateral_magnification_gray (r:gray) = 
                     image_height_gray r /  object_height_genray r`;;


let angular_magnification_genray = define `
    angular_magnification_genray (r:gray) = 
                    image_angle_genray r /  object_angle_genray r`;;

(*----------------------------------------------------------*)
(*------------------Cardianl Points-------------------------*)
(*----------------------------------------------------------*)

(*----------------------------------------------------------*)
(*------------------Principal Points------------------------*)
(*----------------------------------------------------------*)

(*----------------------------------------------------------*)
(*  Note:                                                   *)
(*  sys = composed system, i.e., :(optical_system)list      *)
(*----------------------------------------------------------*)

let principal_planes_specification = define `
    principal_planes_specification sys <=>
     (!ray. (is_valid_gray_in_system ray sys /\ is_valid_genray ray) ==>
            (let M = composed_system sys
             and y_n = image_height_gray ray
             and y_0 = object_height_genray ray in
             ~(y_0 = &0) /\ ~(M$2$1 = (Cx (&0)))
             ==> M$1$2 = (Cx (&0)) /\ lateral_magnification_gray ray = &1))`;;

let principal_ponts = define `
    principal_ponts (dU, dU') sys ni nt = 
    principal_planes_specification
   (gen_optical_system sys dU dU' ni nt)`;;


let pr_ponts_of_system = define `
    pr_ponts_of_system  sys ni nt = 
    @(dU, dU').  principal_ponts (dU, dU') sys ni nt`;;


let lemma_simp = 
    REAL_FIELD ` !A B C D d' d.
    (d' = (&1 - A)/ C  /\ d = D / C * (A - &1) - B)/\ ~(C = &0) ==>
    (A*d + B + C * d * d' + D*d' = &0)`;;

 let lemma_simp2 = 
     REAL_FIELD ` !d' d A B C D.
    (d' = (&1 - A)/ C  /\ d = D / C * (A - &1) - B)/\ ~(C = &0) ==>
    (A + C*d' = &1) `;;


let PRINCIPAL_POINTS = prove(`  
    !sys ni nt. is_valid_optical_system sys /\
    &0 < ni /\ &0 < nt /\
    system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D)  ==>
    let (dU = (D / C) * (A - &1) - B) and
        (dU' = (&1 - A) / C) in 
    principal_ponts (dU, dU') sys ni nt`,

REPEAT LET_TAC THEN 
REPEAT STRIP_TAC THEN
REWRITE_TAC[principal_ponts;principal_planes_specification ] THEN
REPEAT STRIP_TAC THEN 
REPEAT LET_TAC THEN 
REPEAT STRIP_TAC THENL[

UNDISCH_TAC `composed_system (gen_optical_system sys dU dU' ni nt) = M` THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
SIMP_TAC[CMAT2X2_EQ_ALT] THEN
REPEAT STRIP_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[lemma_simp;CX_INJ] THEN
ASM_MESON_TAC[];ALL_TAC] THEN
ASSERT_TAC ` A * dU + B + C * dU * dU' + D * dU' = &0` THENL[

UNDISCH_TAC `composed_system (gen_optical_system sys dU dU' ni nt) = M` THEN 
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN 
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
SIMP_TAC[CMAT2X2_EQ_ALT] THEN
REPEAT STRIP_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[lemma_simp;CX_INJ] THEN
ASM_MESON_TAC[];ALL_TAC] THEN 

REPEAT (POP_ASSUM MP_TAC) THEN
REWRITE_TAC[lateral_magnification_gray;image_height_gray;
           object_height_genray] THEN
REPEAT STRIP_TAC THEN
ASM_REWRITE_TAC[] THEN
MP_TAC COMPOSED_SYSTEM_MATRIX THEN
DISCH_THEN(MP_TAC o SPEC `(gen_optical_system sys dU dU' ni nt)
                            :(optical_system)list`) THEN
DISCH_THEN(MP_TAC o SPEC `(ray : gray)`) THEN
REPEAT LET_TAC THEN
ANTS_TAC THENL[
ASM_SIMP_TAC[is_valid_composed_system;ALL;is_valid_optical_system;
               is_valid_free_space;ALL;gen_optical_system];ALL_TAC] THEN

UNDISCH_TAC `composed_system (gen_optical_system sys dU dU' ni nt) = M` THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
DISCH_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[GSYM cmat2x2;CMAT2X2_VECTOR_MUL;CVECTOR2_EQ] THEN
SIMP_TAC[GSYM CX_MUL;GSYM CX_ADD; CX_INJ; REAL_MUL_LZERO;REAL_ADD_RID] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
REWRITE_TAC[FST] THEN
REPEAT(STRIP_TAC) THEN
POP_ASSUM MP_TAC THEN
POP_ASSUM MP_TAC THEN
ASM_REWRITE_TAC[]THEN
REPEAT(STRIP_TAC) THEN
ASM_REWRITE_TAC[] THEN
ASM_SIMP_TAC[REAL_FIELD`~(y_0 = &0) ==> ( ((A + C * dU') * y_0) / y_0 =  
                                  (A + C * dU'))`] THEN
IMP_REWRITE_TAC[lemma_simp2] THEN
EXISTS_TAC `B:real`THEN
EXISTS_TAC `D:real` THEN
EXISTS_TAC `dU:real` THEN

CONJ_TAC THENL[
ASM_MESON_TAC[];ALL_TAC] THEN
UNDISCH_TAC `M = composed_system (gen_optical_system sys dU dU' ni nt)` THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[CMAT2X2_EQ_ALT] THEN
ASM_MESON_TAC[]);;


let PRINCIPAL_POINTS_ALT = prove (`
    !sys ni nt A B C D.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) /\
        (dU = D / C * (A - &1) - B) /\
        (dU' = (&1 - A) / C) ==>
              principal_ponts (dU,dU') sys ni nt`,
REPEAT GEN_TAC THEN
MP_TAC PRINCIPAL_POINTS THEN 
REPEAT LET_TAC THEN 
MESON_TAC[]);;


let PRINCIPAL_POINTS_AND_ALT = prove(`
 (!sys ni nt.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) /\
        (dU = D / C * (A - &1) - B) /\
        (dU' = (&1 - A) / C) ==>
              principal_ponts (dU,dU') sys ni nt) <=> 
(!sys ni nt. is_valid_optical_system sys /\
    &0 < ni /\ &0 < nt /\
    system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D)  ==>
    let (dU = (D / C) * (A - &1) - B) and
        (dU' = (&1 - A) / C) in 
    principal_ponts (dU, dU') sys ni nt)`,

LET_TAC THEN
REPEAT(POP_ASSUM MP_TAC) THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
ASM_REWRITE_TAC[] THEN
REPEAT STRIP_TAC THEN
EQ_TAC THENL[
MESON_TAC[];ALL_TAC] THEN
REPEAT STRIP_TAC THEN
IMP_REWRITE_TAC[PRINCIPAL_POINTS_ALT]);;


(*----------------------------------------------------------*)
(*------------------Nodal Points ---------------------------*)
(*----------------------------------------------------------*)
let nodal_points_specf = define `
    nodal_points_specf sys <=>
     (!ray. (is_valid_gray_in_system ray sys /\ is_valid_genray ray) ==>
            (let M = composed_system sys
             and y_n = image_height_gray ray
             and y_0 = object_height_genray ray
	     and theta_n = image_angle_genray ray
             and theta_0 = object_angle_genray ray in
             (y_0 = &0) /\ (y_n = &0)/\ ~(theta_0 = &0)/\ ~(M$2$1 = Cx (&0))
             ==> M$1$2 = Cx (&0) /\  angular_magnification_genray ray = &1))`;;



let nodal_points = define `
    nodal_points (dN, dN') sys ni nt = 
    nodal_points_specf
   (gen_optical_system sys dN dN' ni nt)`;;


let lemma_simp3 = 
    REAL_FIELD ` !A B C D d' d.
    (d = (&1 - D)/ C  /\ d' = A / C * (D - &1) - B)/\ ~(C = &0) ==>
    (A*d + B + C * d * d' + D*d' = &0)`;;

 let lemma_simp4 = 
     REAL_FIELD ` !d' d A B C D.
    (d = (&1 - D)/ C  /\ d' = A / C * (D - &1) - B)/\ ~(C = &0) ==>
    (C*d + D = &1) `;;


let NODAL_POINTS = prove(`  
    !sys ni nt. is_valid_optical_system sys /\
    &0 < ni /\ &0 < nt /\
    system_composition sys =  cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) ==>
    let (dN = (&1 - D) / C) and
        (dN' = (A / C) * (D - &1) - B) in 
    nodal_points (dN, dN') sys ni nt`,

REPEAT LET_TAC THEN 
REPEAT STRIP_TAC THEN
REWRITE_TAC[nodal_points;nodal_points_specf ] THEN
REPEAT STRIP_TAC THEN 
REPEAT LET_TAC THEN 
REPEAT STRIP_TAC THENL[

UNDISCH_TAC `composed_system (gen_optical_system sys dN dN' ni nt) = M` THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
SIMP_TAC[CMAT2X2_EQ_ALT;GSYM CX_MUL;GSYM CX_ADD;CX_INJ] THEN
REPEAT STRIP_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[lemma_simp3] THEN
ASM_MESON_TAC[];ALL_TAC] THEN
ASSERT_TAC ` A * dN + B + C * dN * dN' + D * dN' = &0` THENL[

UNDISCH_TAC `composed_system (gen_optical_system sys dN dN' ni nt) = M` THEN 
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN 
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
SIMP_TAC[CMAT2X2_EQ_ALT;GSYM CX_MUL;GSYM CX_ADD;CX_INJ] THEN
REPEAT STRIP_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[lemma_simp3] THEN
ASM_MESON_TAC[];ALL_TAC] THEN 

REPEAT (POP_ASSUM MP_TAC) THEN
REWRITE_TAC[angular_magnification_genray;image_angle_genray;
            object_angle_genray] THEN
REPEAT STRIP_TAC THEN
ASM_REWRITE_TAC[] THEN
MP_TAC COMPOSED_SYSTEM_MATRIX THEN
DISCH_THEN(MP_TAC o SPEC `(gen_optical_system sys dN dN' ni nt)
                            :(optical_system)list`) THEN
DISCH_THEN(MP_TAC o SPEC `(ray : gray)`) THEN
REPEAT LET_TAC THEN
ANTS_TAC THENL[
ASM_SIMP_TAC[is_valid_composed_system;ALL;is_valid_optical_system;
               is_valid_free_space;ALL;gen_optical_system];ALL_TAC] THEN

UNDISCH_TAC `composed_system (gen_optical_system sys dN dN' ni nt) = M` THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
DISCH_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[GSYM cmat2x2;CMAT2X2_VECTOR_MUL;CVECTOR2_EQ;GSYM CX_MUL;GSYM CX_ADD;CX_INJ] THEN
SIMP_TAC[REAL_MUL_LZERO;REAL_ADD_RID] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
REWRITE_TAC[FST] THEN
REPEAT(STRIP_TAC) THEN
POP_ASSUM MP_TAC THEN
POP_ASSUM MP_TAC THEN
ASM_REWRITE_TAC[]THEN
REPEAT(STRIP_TAC) THEN
ASM_REWRITE_TAC[]THEN
UNDISCH_TAC `object_height_genray ray = y_0` THEN
REWRITE_TAC[object_height_genray] THEN 
ASM_REWRITE_TAC[] THEN
SIMP_TAC[REAL_MUL_RZERO;REAL_ADD_LID] THEN DISCH_TAC THEN 
ASM_SIMP_TAC[REAL_FIELD`~(theta_0 = &0) ==> (((C * dN + D) * theta_0) / theta_0 = 
                               (C * dN + D))  `] THEN
IMP_REWRITE_TAC[lemma_simp4] THEN
EXISTS_TAC `A:real`THEN
EXISTS_TAC `B:real` THEN
EXISTS_TAC `dN':real` THEN

CONJ_TAC THENL[
ASM_MESON_TAC[];ALL_TAC] THEN
UNDISCH_TAC `M = composed_system (gen_optical_system sys dN dN' ni nt)` THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[CMAT2X2_EQ_ALT;GSYM CX_MUL;GSYM CX_ADD; CX_INJ] THEN
ASM_MESON_TAC[]);;


let NODAL_POINTS_ALT = prove (`
    !sys ni nt A B C D.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) /\
        (dN = (&1 - D) / C)  /\
        (dN' =  A / C * (D - &1) - B)  ==>
               nodal_points (dN,dN') sys ni nt`,
REPEAT GEN_TAC THEN
MP_TAC NODAL_POINTS THEN 
REPEAT LET_TAC THEN 
MESON_TAC[]);;


(*----------------------------------------------------------*)
(*------------------Focal Points ---------------------------*)
(*----------------------------------------------------------*)

let focal_points_spec = define `
    focal_points_spec sys <=>
     (!ray. (is_valid_gray_in_system ray sys /\ is_valid_genray ray) ==>
            (let M = composed_system sys
             and y_n = image_height_gray ray
             and y_0 = object_height_genray ray
	     and theta_n = image_angle_genray ray
             and theta_0 = object_angle_genray ray in
             ~(M$2$1 = Cx (&0)) ==>
              (((theta_n = &0 /\ y_0 = &0) ==> (M$1$1 = Cx (&0))) /\
               ((theta_0 = &0 /\ y_n = &0) ==> (M$2$2 = Cx (&0))))))`;;



let focal_pts = define `
    focal_pts (dP, dP') sys ni nt = 
    focal_points_spec
   (gen_optical_system sys dP dP' ni nt)`;;


let FOCAL_POINTS = prove(`  
    !sys ni nt. is_valid_optical_system sys /\
    &0 < ni /\ &0 < nt /\
    system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D)  ==>
    let (dF = -- D / C) and
        (dF' = -- A / C) in 
    focal_pts (dF, dF') sys ni nt`,

REPEAT LET_TAC THEN 
REPEAT STRIP_TAC THEN
REWRITE_TAC[focal_pts;focal_points_spec] THEN
REPEAT STRIP_TAC THEN 
REPEAT LET_TAC THEN 
REPEAT STRIP_TAC THENL[

MP_TAC COMPOSED_SYSTEM_MATRIX THEN
DISCH_THEN(MP_TAC o SPEC `(gen_optical_system sys dF dF' ni nt)
                            :(optical_system)list`) THEN
DISCH_THEN(MP_TAC o SPEC `(ray : gray)`) THEN
REPEAT LET_TAC THEN
ANTS_TAC THENL[
ASM_SIMP_TAC[is_valid_composed_system;ALL;is_valid_optical_system;
               is_valid_free_space;ALL;gen_optical_system];ALL_TAC] THEN
UNDISCH_TAC `composed_system (gen_optical_system sys dF dF' ni nt) = M` THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
DISCH_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[GSYM cmat2x2;CMAT2X2_VECTOR_MUL;CVECTOR2_EQ;GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN
REWRITE_TAC[cmat2x2;VECTOR_2;GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN
REPEAT(STRIP_TAC) THEN
IMP_REWRITE_TAC[REAL_FIELD ` 
      ~(C = &0) ==> --A / C = dF' ==> A + C * dF' = &0`] THEN
UNDISCH_TAC `M = composed_system (gen_optical_system sys dF dF' ni nt)` THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[CMAT2X2_EQ_ALT;GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN
ASM_MESON_TAC[]; ALL_TAC] THEN 
MP_TAC COMPOSED_SYSTEM_MATRIX THEN
DISCH_THEN(MP_TAC o SPEC `(gen_optical_system sys dF dF' ni nt)
                            :(optical_system)list`) THEN
DISCH_THEN(MP_TAC o SPEC `(ray : gray)`) THEN
REPEAT LET_TAC THEN
ANTS_TAC THENL[
ASM_SIMP_TAC[is_valid_composed_system;ALL;is_valid_optical_system;
               is_valid_free_space;ALL;gen_optical_system];ALL_TAC] THEN
UNDISCH_TAC `composed_system (gen_optical_system sys dF dF' ni nt) = M` THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
DISCH_TAC THEN
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[GSYM cmat2x2;CMAT2X2_VECTOR_MUL;CVECTOR2_EQ;GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN
REWRITE_TAC[cmat2x2;VECTOR_2;GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN
REPEAT(STRIP_TAC) THEN
IMP_REWRITE_TAC[REAL_FIELD ` 
   ~(C = &0) ==> --D / C = dF ==> C * dF + D = &0`] THEN
UNDISCH_TAC `M = composed_system (gen_optical_system sys dF dF' ni nt)` THEN
IMP_REWRITE_TAC[ GEN_OPTICAL_SYSTEM_MAT] THEN
SIMP_TAC[CMAT2X2_EQ_ALT;GSYM CX_ADD;GSYM CX_MUL;CX_INJ] THEN
ASM_MESON_TAC[]);;


let FOCAL_POINTS_ALT = prove (`
    !sys ni nt A B C D.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) /\
        (dF = -- D / C) /\
        (dF' = -- A / C) ==>
         focal_pts (dF,dF') sys ni nt`,
REPEAT GEN_TAC THEN
MP_TAC FOCAL_POINTS THEN 
REPEAT LET_TAC THEN 
MESON_TAC[]);;




let CARDINAL_POINTS = prove(`!sys ni nt A B C D.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\
         system_composition sys =  cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) ==> 
        (let  (dU = D / C * (A - &1) - B) and 
              (dU' = (&1 - A) / C) and
	      (dN = (&1 - D) / C) and
	      (dN' = A / C * (D - &1) - B) and 
	      (dF = --D / C) and
	      (dF' = --A / C) in 
         (principal_ponts (dU,dU') sys ni nt /\
	  nodal_points (dN,dN') sys ni nt /\
	  focal_pts (dF,dF') sys ni nt))`,
REPEAT GEN_TAC THEN
MP_TAC PRINCIPAL_POINTS THEN 
MP_TAC NODAL_POINTS THEN 
MP_TAC FOCAL_POINTS THEN
REPEAT LET_TAC THEN 
MESON_TAC[]);;


(*----------------------------------------------------------*)
(*------------------Focal Length ---------------------------*)
(*----------------------------------------------------------*)

let focal_length_obj_space = define `
    focal_length_obj_space FL sys ni nt <=> (?dU dF dU' dF'. 
    focal_pts (dF,dF') sys ni nt /\ principal_ponts (dU,dU') sys ni nt /\
    (FL = dU - dF)) `;;

let focal_length_image_space = define `
    focal_length_image_space FL' sys ni nt <=> (?dU dF dU' dF'. 
    focal_pts (dF,dF') sys ni nt /\ principal_ponts (dU,dU') sys ni nt /\
    (FL' = dF' - dU')) `;;


let FOCAL_LENGTH_OBJECT_SPACE = prove(`
         !sys ni nt.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\ ~(C = &0) /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) ==>
	 focal_length_obj_space  ((A*D - B*C) / C) sys ni nt`,

REPEAT STRIP_TAC THEN
REWRITE_TAC[ focal_length_obj_space] THEN
EXISTS_TAC `D / C * (A - &1) - B` THEN
EXISTS_TAC ` --D / (C:real)` THEN
EXISTS_TAC `(&1 - A) / C ` THEN
EXISTS_TAC ` --A / (C:real) ` THEN
IMP_REWRITE_TAC[FOCAL_POINTS_ALT;PRINCIPAL_POINTS_ALT] THEN
ASM_SIMP_TAC[REAL_FIELD ` ~(C = &0) ==> (A * D - B * C) / C = 
              D / C * (A - &1) - B - --D / C`]);;


let FOCAL_LENGTH_IMAGE_SPACE = prove(`
         !sys ni nt.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\ ~(C = &0) /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) ==>
	 focal_length_image_space  (-- (&1 / C)) sys ni nt`,

REPEAT STRIP_TAC THEN
REWRITE_TAC[ focal_length_image_space] THEN
EXISTS_TAC `D / C * (A - &1) - B` THEN
EXISTS_TAC ` --D / (C:real)` THEN
EXISTS_TAC `(&1 - A) / C ` THEN
EXISTS_TAC ` --A / (C:real) ` THEN
IMP_REWRITE_TAC[FOCAL_POINTS_ALT;PRINCIPAL_POINTS_ALT] THEN
asm_real_prove);;


let eye_corrective_system = new_definition`
    eye_corrective_system n0 n1 ZG  AD BD CD DD A B C D = 
    (let device = unknown AD BD CD DD
     and eye_model = unknown A B C D in
     [(n0,&0),device; (n0,ZG),eye_model],(n1,&0))`;;


let VISUAL_OPTICAL_SYSTEMS = prove (`
!n0 n1 AD BD CD DD A B C D ZG.
 &0 < n0 /\
 &0 < n1 /\
system_composition (eye_corrective_system n0 n1 ZG  AD BD CD DD A B C D) = 
cmat2x2 (Cx Ac) (Cx Bc) (Cx Cc) (Cx Dc) ==>

(let dU = Dc / Cc * (Ac - &1) - Bc
 and dU' = (&1 - Ac) / Cc
 and dN = (&1 - Dc) / Cc
 and dN' = Ac / Cc * (Dc - &1) - Bc
 and dF = --Dc / Cc
 and dF' = --Ac / Cc in
     principal_ponts (dU,dU') 
       (eye_corrective_system n0 n1 ZG  AD BD CD DD A B C D) n0 n1 /\
     nodal_points (dN,dN') 
       (eye_corrective_system n0 n1 ZG  AD BD CD DD A B C D) n0 n1 /\
     focal_pts (dF,dF')
       (eye_corrective_system n0 n1 ZG  AD BD CD DD A B C D) n0 n1)`,

IMP_REWRITE_TAC[CARDINAL_POINTS] THEN 
SIMP_TAC[eye_corrective_system; is_valid_optical_system] THEN 
REPEAT (LET_TAC) THEN 
REPEAT(POP_ASSUM MP_TAC) THEN 
SIMP_TAC[EQ_SYM_EQ] THEN 
SIMP_TAC[is_valid_optical_system;ALL;is_valid_optical_component;
         is_valid_free_space;is_valid_interface]);;




