(* ========================================================================= *)
(*                                                                           *)
(*               Component Libray for Geometrical Optical Analysis           *)
(*                                                                           *)
(*        (c) Copyright, Umair Siddiquen=, 2012-2015                         *)
(*                       Hardware Verification Group (HVG),                  *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last update: September 10,2015                                            *)
(*                                                                           *)
(* ========================================================================= *)

needs "cardinal_points.ml";;


(*-----THIN LENS------------*)

let thin_lens_mat = new_definition 
  `thin_lens_mat R1 R2 (n1:real) (n2:real) = 
     cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(((n2 - n1)/n1) *(&1/R2 - &1/R1))) (Cx(&1)) `;;

let thin_lens = new_definition
  `thin_lens R1 R2 n1 n2: optical_system =
    ([(n1,&0),(spherical_t R1);
      (n2,&0),(spherical_t R2)],n1,&0)`;;

let VALID_OPTICAL_SYSTEM_TAC x =
  SIMP_TAC[x;is_valid_optical_system;is_valid_optical_component;ALL;
    is_valid_free_space;is_valid_interface]
  THEN CONV_TAC REAL_FIELD;;

let VALID_THIN_LENS = prove
  (`!R1 R2 n1 n2.
    ~(R1= &0) /\ ~(R2= &0) /\ &0 < n1 /\ &0 < n2 ==>
      is_valid_optical_system (thin_lens R1 R2 n1 n2)`,
  VALID_OPTICAL_SYSTEM_TAC thin_lens);;

let THIN_LENS_THIN_LENS_MAT = prove
  (`!R1 R2 n1 n2.
    ~(R1 = &0) /\ ~(R2 = &0) /\ &0 < n1 /\ &0 < n2 ==>
      system_composition (thin_lens R1 R2 n1 n2) = thin_lens_mat R1 R2 n1 n2`,
  REWRITE_TAC[system_composition;head_index;interface_matrix;free_space_matrix;
    thin_lens_mat;thin_lens;CMAT2X2_MUL;CMAT2X2_EQ;GSYM CX_ADD;GSYM CX_MUL;CX_INJ]
  THEN CONV_TAC REAL_FIELD);;

let THIN_LENS_MATRIX = prove
  (`!R1 R2 n1 n2.
    ~(R1= &0) /\ ~(R2= &0) /\ &0 < n1 /\ &0 < n2 ==>
    !ray. is_valid_ray_in_system ray (thin_lens R1 R2 n1 n2) ==> 
    let ((y0,theta0),(y1,theta1),rs) = ray in
    let (y_n,theta_n) = last_single_ray ray in
   vector [Cx y_n;Cx theta_n] = thin_lens_mat R1 R2 n1 n2 ** 
   vector [Cx y0;Cx theta0]`,
   SIMP_TAC[GSYM THIN_LENS_THIN_LENS_MAT] THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC SYSTEM_MATRIX THEN ASM_SIMP_TAC[VALID_THIN_LENS]);;

(*------------------THICK LENS---------------------*)
let thick_lens_mat = new_definition 
  `thick_lens_mat R1 R2 (n1:real) (n2:real) d = 
     cmat2x2 (Cx (&1 + d*(n1/(n2*R1) - &1/R1))) (Cx (d * n1 / n2))
            (Cx (--(((n1 - n2)*(d*(n1-n2) + n2*(R1-R2)))/(n1*n2*R1*R2))))
            (Cx (&1 + d*(&1/R2 - n1/(n2*R2))))   `;;

let thick_lens = new_definition
  `thick_lens R1 R2 n1 n2 d: optical_system =
    ([(n1,&0),(spherical_t R1);
      (n2,d),(spherical_t R2)],n1,&0)`;;


let VALID_THICK_LENS = prove
  (`!R1 R2 n1 n2.
    ~(R1= &0) /\ ~(R2= &0) /\ &0 < n1 /\ &0 < n2 /\ &0 <= d ==>
      is_valid_optical_system (thick_lens R1 R2 n1 n2 d)`,
  VALID_OPTICAL_SYSTEM_TAC thick_lens);;

let THICK_LENS_THICK_LENS_MAT = prove
  (`!R1 R2 n1 n2 d.
    ~(R1 = &0) /\ ~(R2 = &0) /\ &0 < n1 /\ &0 < n2 /\ &0 <= d ==>
     system_composition (thick_lens R1 R2 n1 n2 d) = thick_lens_mat R1 R2 n1 n2 d`,
  REWRITE_TAC[system_composition;head_index;interface_matrix;free_space_matrix;
    thick_lens_mat;thick_lens;CMAT2X2_MUL;CMAT2X2_EQ;GSYM CX_ADD;GSYM CX_MUL;CX_INJ]
  THEN CONV_TAC REAL_FIELD);;

let THICK_LENS_MATRIX = prove
  (`!R1 R2 n1 n2.
    ~(R1= &0) /\ ~(R2= &0) /\ &0 < n1 /\ &0 < n2 /\ &0 <= d ==>
    !ray. is_valid_ray_in_system ray (thick_lens R1 R2 n1 n2 d) ==> 
    let ((y0,theta0),(y1,theta1),rs) = ray in
    let (y_n,theta_n) = last_single_ray ray in
   vector [Cx y_n;Cx theta_n] = thick_lens_mat R1 R2 n1 n2 d ** 
   vector [Cx y0;Cx theta0]`,
   SIMP_TAC[GSYM THICK_LENS_THICK_LENS_MAT] THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC SYSTEM_MATRIX THEN ASM_SIMP_TAC[VALID_THICK_LENS]);;


let DET_MUL3 = prove (` 
               !A B C. det (A ** B ** c) = det A * det B * det c`, 
               MESON_TAC[DET_MUL]);;

let THICK_LENS_MATRIX_PROD = prove (
    ` !R1 R2 n1 n2.
    ~(R1= &0) /\ ~(R2= &0) /\ &0 < n1 /\ &0 < n2 /\ &0 <= d ==>
     thick_lens_mat R1 R2 n1 n2 d = 
     cmat2x2 (Cx (&1)) (Cx(&0)) (Cx((n2 - n1) / n1 * &1 / R2)) (Cx(n2 / n1)) **
     cmat2x2 (Cx (&1)) (Cx d) (Cx (&0)) (Cx (&1)) **
     cmat2x2 (Cx(&1)) (Cx(&0)) (Cx((n1 - n2) / n2 * &1 / R1)) (Cx(n1 / n2))` ,
    
     REWRITE_TAC[system_composition; head_index; interface_matrix;
                 free_space_matrix;thick_lens_mat;thick_lens;
                 CMAT2X2_MUL;CMAT2X2_EQ;GSYM CX_ADD;GSYM CX_MUL;CX_INJ]
     THEN CONV_TAC REAL_FIELD );;



(*-----------DIELECTRIC SLAB---------------*)

let slab_mat = new_definition 
  `slab_mat (n1:real) (n2:real) d = 
     cmat2x2 (Cx(&1)) (Cx(d*(n1/n2))) (Cx(&0)) (Cx(&1)) `;;

let slab = new_definition
  `slab n1 n2 d: optical_system =
    ([(n1,&0),(plane_t);
      (n2,d),(plane_t)],n1,&0)`;;


let VALID_SLAB = prove
  (`!n1 n2 d.
    &0 < n1 /\ &0 < n2 /\   &0 < d ==>
      is_valid_optical_system (slab n1 n2 d)`,
  VALID_OPTICAL_SYSTEM_TAC slab);;

let SLAB_SLAB_MAT = prove
  (` !n1 n2 d.
     &0 < n1 /\ &0 < n2  /\ &0 < d ==>
      system_composition (slab n1 n2 d) = (slab_mat n1 n2 d)`,
  REWRITE_TAC[system_composition;head_index;interface_matrix;free_space_matrix;
    slab_mat;slab;CMAT2X2_MUL;CMAT2X2_EQ;GSYM CX_ADD;GSYM CX_MUL;CX_INJ]
  THEN CONV_TAC REAL_FIELD);;

let SLAB_MATRIX = prove
  (`!n1 n2 d.
    &0 < n1 /\ &0 < n2 /\ &0 < d ==>
    !ray. is_valid_ray_in_system ray (slab n1 n2 d) ==> 
    let ((y0,theta0),(y1,theta1),rs) = ray in
    let (y_n,theta_n) = last_single_ray ray in
   vector [Cx y_n;Cx theta_n] = slab_mat n1 n2 d ** vector [Cx y0;Cx theta0]`,
   SIMP_TAC[GSYM SLAB_SLAB_MAT] THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC SYSTEM_MATRIX THEN ASM_SIMP_TAC[VALID_SLAB]);;


(*-------------------------------------------------------------------*)
(*-----------VERIFICATION OF CARDINAL POINTS ------------------------*)
(*-------------------------------------------------------------------*)


(*----------------------------------------------------------*)
(*---Cardinal Points of Spherical Transmitted Interface-----*)
(*----------------------------------------------------------*)

let spherical_transm = define `
        spherical_transm  n n' R = 
        (([(n,&0),spherical_t R],(n',&0)):optical_system)`;;


let SPHERICAL_TRANS_MATRIX = prove(
` !n n' R. system_composition (spherical_transm  n n' R) = 
           interface_matrix n n' (spherical_t R)`,

SIMP_TAC[spherical_transm ;system_composition; 
         interface_matrix;free_space_matrix] THEN
SIMP_TAC[head_index;CMAT2X2_LID;CMAT2X2_RID]);;


let PRINCIPAL_POINTS_SI_TRANS = prove(
   `!R n n'.  ~(R = &0) /\  &0 < n /\ &0 < n'  ==>
   (let dU = &0 and 
        dU' = &0 in 
    principal_ponts (dU,dU') (spherical_transm  n n' R) n n')`,

REPEAT STRIP_TAC THEN
REPEAT LET_TAC THEN
MATCH_MP_TAC PRINCIPAL_POINTS_ALT THEN
EXISTS_TAC `&1` THEN 
EXISTS_TAC `&0` THEN 
EXISTS_TAC `((n - n') / n' * &1 / R)` THEN 
EXISTS_TAC `((n:real) / n')` THEN 
SIMP_TAC[ SPHERICAL_TRANS_MATRIX] THEN
SIMP_TAC[interface_matrix;GSYM CX_ADD;GSYM CX_MUL;CX_INJ;VECTOR_2] THEN
ASM_SIMP_TAC[is_valid_optical_system;spherical_transm;ALL;
             is_valid_optical_component;  is_valid_free_space;
             is_valid_interface ] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
CONV_TAC REAL_FIELD);;


let NODAL_POINTS_SI_TRANS = prove(
   `!R n n'.  ~(R = &0) /\  &0 < n /\ &0 < n' /\ ~(n' = n)  ==>
   (let dN = --  R and 
        dN' =  R in 
    nodal_points (dN,dN') (spherical_transm  n n' R) n n')`,

REPEAT STRIP_TAC THEN
REPEAT LET_TAC THEN
MATCH_MP_TAC NODAL_POINTS_ALT THEN
EXISTS_TAC `&1` THEN 
EXISTS_TAC `&0` THEN 
EXISTS_TAC `((n - n') / n' * &1 / R)` THEN 
EXISTS_TAC `((n:real) / n')`  THEN
SIMP_TAC[ SPHERICAL_TRANS_MATRIX] THEN
SIMP_TAC[interface_matrix;VECTOR_2] THEN
ASM_SIMP_TAC[is_valid_optical_system;spherical_transm;ALL;
             is_valid_optical_component;  is_valid_free_space;
             is_valid_interface ] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
CONV_TAC REAL_FIELD);;


let FOCAL_POINTS_SI_TRANS = prove(
   `!R n n'.  ~(R = &0) /\  &0 < n /\ &0 < n' /\ ~(n' = n)  ==>
   (let dF =  (n * R) / (n' - n) and 
        dF' =  (n' * R) / (n' - n) in 
      focal_pts  (dF,dF') (spherical_transm  n n' R) n n')`,

REPEAT STRIP_TAC THEN
REPEAT LET_TAC THEN
MATCH_MP_TAC FOCAL_POINTS_ALT THEN
EXISTS_TAC `(&1)` THEN
EXISTS_TAC `(&0)` THEN
EXISTS_TAC `((n - n') / n' * &1 / R)` THEN
EXISTS_TAC `((n:real) / n')` THEN 

SIMP_TAC[interface_matrix; SPHERICAL_TRANS_MATRIX] THEN
SIMP_TAC[interface_matrix;VECTOR_2] THEN
ASM_SIMP_TAC[is_valid_optical_system;spherical_transm;ALL;
             is_valid_optical_component;  is_valid_free_space;
             is_valid_interface ] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
CONV_TAC REAL_FIELD);;

(*----------------------------------------------------------*)
(*---Cardinal Points of Spherical REFLECTED Interface-------*)
(*----------------------------------------------------------*)

let spherical_reflected = define `
        spherical_reflected  n n' R = 
        (([(n,&0),spherical_r R],(n',&0)):optical_system)`;;


let SPHERICAL_REFL_MATRIX = prove(
` !n n' R. system_composition (spherical_reflected  n n' R) = 
           interface_matrix n n' (spherical_r R) `,

SIMP_TAC[spherical_reflected ;system_composition; 
         interface_matrix;free_space_matrix] THEN
SIMP_TAC[head_index;CMAT2X2_LID;CMAT2X2_RID;GSYM CX_ADD;GSYM CX_MUL;CX_INJ]);;


let PRINCIPAL_POINTS_SI_REFL = prove(
   `!R n n'.  ~(R = &0) /\  &0 < n /\ &0 < n'  ==>
   (let dU = &0 and 
        dU' = &0 in 
    principal_ponts (dU,dU') (spherical_reflected  n n' R) n n')`,

REPEAT STRIP_TAC THEN
REPEAT LET_TAC THEN
MATCH_MP_TAC PRINCIPAL_POINTS_ALT THEN
EXISTS_TAC `&1` THEN 
EXISTS_TAC `&0` THEN 
EXISTS_TAC `(-- &2 / R)` THEN 
EXISTS_TAC `&1` THEN 
SIMP_TAC[ SPHERICAL_REFL_MATRIX] THEN
SIMP_TAC[interface_matrix;VECTOR_2] THEN
ASM_SIMP_TAC[is_valid_optical_system;spherical_reflected;ALL;
             is_valid_optical_component;  is_valid_free_space;
             is_valid_interface ] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
CONV_TAC REAL_FIELD);;


let NODAL_POINTS_SI_REFL = prove(
   `!R n n'.  ~(R = &0) /\  &0 < n /\ &0 < n' /\ ~(n' = n)  ==>
   (let dN = &0 and 
        dN' =  &0 in 
    nodal_points (dN,dN') (spherical_reflected  n n' R) n n')`,

REPEAT STRIP_TAC THEN
REPEAT LET_TAC THEN
MATCH_MP_TAC NODAL_POINTS_ALT THEN
EXISTS_TAC `&1` THEN 
EXISTS_TAC `&0` THEN 
EXISTS_TAC `(-- &2 / R)` THEN 
EXISTS_TAC `&1` THEN 
SIMP_TAC[ SPHERICAL_REFL_MATRIX] THEN
SIMP_TAC[interface_matrix;VECTOR_2] THEN
ASM_SIMP_TAC[is_valid_optical_system;spherical_reflected;ALL;
             is_valid_optical_component;  is_valid_free_space;
             is_valid_interface ] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
CONV_TAC REAL_FIELD);;

(*---------For CONVEX mirror -----------------*)

let FOCAL_POINTS_SI_REFL = prove(
   `!R n n'.  ~(R = &0) /\  &0 < n /\ &0 < n' /\ ~(n' = n)  ==>
   (let dF =   R / &2 and 
        dF' =  R / &2 in 
      focal_pts  (dF,dF') (spherical_reflected  n n' R) n n')`,

REPEAT STRIP_TAC THEN
REPEAT LET_TAC THEN
MATCH_MP_TAC FOCAL_POINTS_ALT THEN
EXISTS_TAC `(&1)` THEN
EXISTS_TAC `(&0)` THEN
EXISTS_TAC `( -- &2 / R)` THEN
EXISTS_TAC `(&1)` THEN 

SIMP_TAC[interface_matrix; SPHERICAL_REFL_MATRIX] THEN
SIMP_TAC[interface_matrix;VECTOR_2] THEN
ASM_SIMP_TAC[is_valid_optical_system;spherical_reflected;ALL;
             is_valid_optical_component;  is_valid_free_space;
             is_valid_interface ] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
CONV_TAC REAL_FIELD);;


(*----------------------------------------------------------*)
(*-----------------------THICK LENS-------------------------*)
(*----------------------------------------------------------*)

let thick_lens_lemma = 
    REAL_FIELD` ~(R1 = &0) /\ ~(R2 = &0) /\ &0 < n1 /\ &0 < n2 /\ &0 <= d ==> 
    --(((n1 - n2) * (d * (n1 - n2) + n2 * (R1 - R2))) / (n1 * n2 * R1 * R2)) =
    --((n2 - n1) / n1 * (&1 / R1 - &1 / R2 + (((n2 - n1) / n2) * 
      (d / (R1 * R2)))))`;;

let REAL_DIV_DIV = 
    REAL_FIELD ` ! a b c. ~(c = &0) ==> a / (b / c) = (c*a) / b`;;

let thick_lens_lemma2 = 
    REAL_FIELD `!A B C D.  ~(C = &0)  /\ (A*D - B*C = &1) ==> 
    (D / C) * (A - &1) - B = (&1 - D) / C `;;


let PRINCIPAL_POINTS_DET_1 = prove
 (`! A B C D sys ni nt M.
         is_valid_optical_system sys /\
         &0 < ni /\
         &0 < nt /\ ~(C = &0)  /\ (A* D - B * C  = &1) /\
         system_composition sys = cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) /\
         dU = (&1 - D) / C /\ 
         dU' = (&1 - A) / C
         ==> principal_ponts (dU,dU') sys ni nt`,
  REPEAT STRIP_TAC THEN IMP_REWRITE_TAC [PRINCIPAL_POINTS_ALT] THEN
SIMP_TAC[CMAT2X2_EQ] THEN 
  IMP_REWRITE_TAC [thick_lens_lemma2]);;




let PRINCIPAL_POINTS_THICK_LENS = prove
 (`!R1 R2 n nL d.  ~(R1 = &0) /\ ~(R2 = &0) /\ &0 < n /\ &0 < nL /\ &0 <= d  /\
      ~(n = nL) /\  ~(d * (n - nL) =  --nL * (R1 - R2))    ==>
   (let dU =  (n*d*R1) / (nL * (R2 - R1)  + (nL - n)*d)    and 
        dU' = --(n*d*R2) / (nL * (R2 - R1)  + (nL - n)*d) in 
    principal_ponts (dU,dU') (thick_lens R1 R2 n nL d) n n )`,

REPEAT STRIP_TAC THEN REPEAT LET_TAC THEN
MATCH_MP_TAC PRINCIPAL_POINTS_DET_1 THEN
EXISTS_TAC `(&1 + d * (n / (nL * R1) - &1 / R1))` THEN 
EXISTS_TAC `((d:real) * n / nL)` THEN 
EXISTS_TAC `(--(((n - nL) * (d * (n - nL) + nL * (R1 - R2))) /
             ((n:real) * nL * R1 * R2)))` THEN 
EXISTS_TAC `(&1 + d * (&1 / R2 - n / (nL * R2)))` THEN
ASM_SIMP_TAC [GSYM thick_lens_mat;THICK_LENS_THICK_LENS_MAT] THEN 
IMP_REWRITE_TAC [VALID_THICK_LENS] THEN
REPEAT(POP_ASSUM MP_TAC) THEN 
SIMP_TAC[EQ_SYM_EQ] THEN 
CONV_TAC REAL_FIELD );;





let PRINCIPAL_POINTS_BALL_LENS = prove
 (`!R d n nL.  (&0 < R) /\ &0 < n /\ &0 < nL /\
      ~(n = nL)   ==>
   (let dU =  --R    and 
        dU' = --R in 
    principal_ponts (dU,dU') (thick_lens R (--R) n nL (&2*R)) n n )`,
REPEAT STRIP_TAC THEN REPEAT LET_TAC THEN
MATCH_MP_TAC PRINCIPAL_POINTS_DET_1 THEN 
IMP_REWRITE_TAC [VALID_THICK_LENS] THEN 
POP_ASSUM MP_TAC THEN 
POP_ASSUM MP_TAC THEN 
REPEAT STRIP_TAC THEN
EXISTS_TAC `(&1 +  (&2 * R) * (n / (nL * R) - &1 / R))` THEN 
EXISTS_TAC `( (&2 * R) * n / nL)` THEN 
EXISTS_TAC `(--(((n - nL) * ( (&2 * R) * (n - nL) + nL * (R - (--R)))) /
             (n * (nL:real)* R * (--R))))` THEN 
EXISTS_TAC `(&1 +  (&2 * R) * (&1 / (--R) - n / (nL * (--R))))` THEN
SIMP_TAC [GSYM thick_lens_mat] THEN 
IMP_REWRITE_TAC[THICK_LENS_THICK_LENS_MAT] THEN
REPEAT(POP_ASSUM MP_TAC) THEN 
CONV_TAC REAL_FIELD );;


let PRINCIPAL_POINTS_MENISCUS_LENS = prove
 (`!R nL d.  ~(R = &0)  /\ &0 < nL /\ &0 < d  /\
      ~(&1 = nL)     ==>
   (let dU = (R) / (nL  - &1)    and 
        dU' = --R / (nL - &1) in 
    principal_ponts (dU,dU') (thick_lens R R (&1) nL d) (&1) (&1) )`,
REPEAT STRIP_TAC THEN REPEAT LET_TAC THEN
MATCH_MP_TAC PRINCIPAL_POINTS_DET_1 THEN 
IMP_REWRITE_TAC [VALID_THICK_LENS] THEN 
EXISTS_TAC `(&1 + d * (&1 / (nL * R) - &1 / R))` THEN 
EXISTS_TAC `(d * (&1) / nL)` THEN 
EXISTS_TAC `(--(((&1 - nL) * (d * (&1 - nL) + nL * (R - R))) /
             ((&1) * nL * R * R)))` THEN 
EXISTS_TAC `(&1 + d * (&1 / R - (&1) / (nL * R)))` THEN
SIMP_TAC [GSYM thick_lens_mat] THEN 
IMP_REWRITE_TAC[THICK_LENS_THICK_LENS_MAT] THEN 
REPEAT(POP_ASSUM MP_TAC)
THEN 
CONV_TAC REAL_FIELD );;


