(* ========================================================================= *)
(*                                                                           *)
(*                        Formalization of Ray Optics                        *)
(*                                                                           *)
(*        (c) Copyright, Umair Siddique, Vincent Aravantinos, 2012 - 2013    *)
(*                       Umair Siddique 2012-2015                            *)
(*                       Hardware Verification Group (HVG),                  *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last update: September 10,2015                                            *)
(*                                                                           *)
(* ========================================================================= *)

needs "cmat2x2.ml";;

(* ------------------------------------------------------------------------- *)

(* Preliminaries                                                             *)
(* ------------------------------------------------------------------------- *)

let LET_PAIR = prove
  (`!P t. (let x,y = t in P x y) = P (FST t) (SND t)`,
  ONCE_REWRITE_TAC[GSYM PAIR] THEN PURE_REWRITE_TAC[LET_DEF;LET_END_DEF;FST;SND]
  THEN REWRITE_TAC[]);;

let ALL_REVERSE = prove
  (`!l P. ALL P (REVERSE l) = ALL P l`,
  LIST_INDUCT_TAC THEN ASM_REWRITE_TAC[REVERSE;ALL;ALL_APPEND] THEN ITAUT_TAC);;



let BOUNDED_MATRIX_BOUNDED_VECTOR_MUL = prove
  (`!M:num->real^M^N.
      (?b. !n i j. 1 <= i /\ i <= dimindex(:N) /\ 1 <= j /\ j <= dimindex(:M)
          ==> abs ((M n)$i$j) <= b)
      ==> (!v:real^M.
        ?b. !n i. 1 <= i /\ i <= dimindex(:N) ==> abs ((M n ** v)$i) <= b)`,
  REPEAT STRIP_TAC
  THEN SIMP_TAC[MATRIX_VECTOR_MUL_COMPONENT;dot]
  THEN Pa.EXISTS_TAC "sum (1..dimindex(:M)) (\i. b * abs (v$i))"
  THEN REPEAT GEN_TAC THEN DISCH_TAC THEN MATCH_MP_TAC SUM_ABS_LE
  THEN REWRITE_TAC[FINITE_NUMSEG;IN_NUMSEG;REAL_ABS_MUL]
  THEN GEN_TAC THEN Pa.ASM_CASES_TAC "v$x = &0"
  THEN ASM_REWRITE_TAC[REAL_ABS_0;REAL_MUL_RZERO;REAL_LE_REFL]
  THEN MP_REWRITE_TAC REAL_LE_RMUL_EQ
  THEN ASM_MESON_TAC[REAL_LT_LE;REAL_ABS_ZERO;REAL_ABS_POS]);;


(* ------------------------------------------------------------------------- *)
(* System formalization                                                      *)
(* ------------------------------------------------------------------------- *)

new_type_abbrev ("free_space",`:real#real`);;

let FORALL_FREE_SPACE_THM = prove
  (`!P. (!fs:free_space. P fs) <=> !n d. P (n,d)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

let EXISTS_FREE_SPACE_THM = prove
  (`!P. (?fs:free_space. P fs) <=> ?n d. P (n,d)`,
  REWRITE_TAC[EXISTS_PAIR_THM]);;



let interface_IND,interface_REC = define_type
  "interface = plane_t | plane_r | 
               spherical_t real | spherical_r real |
               pcm | unknown complex complex complex complex ";;



new_type_abbrev("optical_component",`:free_space#interface`);; 

let FORALL_OPTICAL_COMPONENT_THM = prove
  (`!P. (!c:optical_component. P c) <=> !fs i ik. P (fs,i)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

let EXISTS_OPTICAL_COMPONENT_THM = prove
  (`!P. (?c:optical_component. P c) <=> ?fs i. P (fs,i)`,
  REWRITE_TAC[EXISTS_PAIR_THM]);;

new_type_abbrev("optical_system",`:optical_component list#free_space`);;

let FORALL_OPTICAL_SYSTEM_THM = prove
  (`!P. (!os:optical_system. P os) <=> !cs fs. P (cs,fs)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

let EXISTS_OPTICAL_SYSTEM_THM = prove
  (`!P. (?os:optical_system. P os) <=> ?cs fs. P (cs,fs)`,
  REWRITE_TAC[EXISTS_PAIR_THM]);;

let is_valid_free_space = define
  `is_valid_free_space ((n,d):free_space) <=> &0 < n `;;

let is_valid_interface = define
  `(is_valid_interface (plane_t) <=> T) /\
   (is_valid_interface (plane_r) <=> T) /\
   (is_valid_interface (spherical_t R) <=> ~(&0 = R))/\ 
   (is_valid_interface (spherical_r R) <=> ~(&0 = R))/\
   (is_valid_interface (pcm) <=> T) /\
   (is_valid_interface (unknown a b c d) <=> T)`;;



let is_valid_optical_component = new_definition
  `is_valid_optical_component (fs,i) <=>
    is_valid_free_space fs /\ is_valid_interface i`;;

let is_valid_optical_system = new_definition
  `is_valid_optical_system (cs,fs) <=>
    ALL is_valid_optical_component cs /\ is_valid_free_space fs`;;

let head_index = define
  `head_index ([],(n,d)) = n /\ 
  head_index (CONS ((n,d),i) cs, (nt,dt)) = n`;;

let HEAD_INDEX_VALID_OPTICAL_SYSTEM = prove
  (`!os:optical_system. is_valid_optical_system os ==> &0 < head_index os`,
  REWRITE_TAC[FORALL_OPTICAL_SYSTEM_THM;is_valid_optical_system]
  THEN MATCH_MP_TAC list_INDUCT
  THEN SIMP_TAC[FORALL_OPTICAL_COMPONENT_THM;FORALL_FREE_SPACE_THM;
    head_index;is_valid_free_space;ALL;is_valid_optical_component]);;

let HEAD_INDEX_APPEND = prove
  (`!cs1:optical_component list cs2 fs a. 
      head_index (APPEND cs1 cs2, fs)
        = head_index (cs1, (head_index (cs2,fs),a))`,
    MATCH_MP_TAC list_INDUCT
    THEN REWRITE_TAC[APPEND;head_index;FORALL_OPTICAL_COMPONENT_THM;
      FORALL_FREE_SPACE_THM]);;

let HEAD_INDEX_CONS_ALT = prove
  (`!fs i cs:optical_component list fs'.
      head_index (CONS (fs,i) cs,fs')
        = head_index ([]:optical_component list,fs)`,
  REWRITE_TAC[FORALL_FREE_SPACE_THM;head_index]);;

let HEAD_INDEX_CONS_EQ = prove
  (`!n1 n2 d1 d2 i1 i2 cs1 cs2 fs1 fs2.
      head_index (CONS ((n1,d1),i1) cs1,fs1)
        = head_index (CONS ((n2,d2),i2) cs2,fs2)
          <=> n1 = n2`,
  REWRITE_TAC[FORALL_PAIR_THM;head_index;PAIR_EQ]);;
  
let HEAD_INDEX_CONS_EQ_ALT = prove
  (`!c1 c2 cs1 cs2 fs1 fs2.
      c1 = c2 ==> head_index (CONS c1 cs1,fs1) = head_index (CONS c2 cs2,fs2)`,
  SIMP_TAC[FORALL_PAIR_THM;head_index;PAIR_EQ]);;


(* ------------------------------------------------------------------------- *)
(* Ray formalization                                                         *)
(* ------------------------------------------------------------------------- *)

new_type_abbrev("single_ray",`:real#real`);;

let FORALL_SINGLE_RAY_THM = prove
  (`!P. (!sr:single_ray. P sr) <=> !y theta. P (y,theta)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

new_type_abbrev("ray",
  `:single_ray # single_ray # (single_ray # single_ray) list`);;

let FORALL_RAY_THM = prove
  (`!P. (!r:ray. P r) <=> !sr1 sr2 srs. P (sr1,sr2,srs)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

let is_valid_ray_in_free_space = define 
  `is_valid_ray_in_free_space (y0,theta0) (y1,theta1) ((n0,d):free_space) <=>
    y1 = y0 + d * theta0 /\ theta0 = theta1`;;

let is_valid_ray_at_interface = define 
  `(is_valid_ray_at_interface 
   (y0,theta0) (y1,theta1) n0 n1 plane_t
    <=> y1 = y0 /\ n0 * theta0 =  n1 * theta1) /\ 
   
   (is_valid_ray_at_interface 
   (y0,theta0) (y1,theta1) n0 n1 plane_r <=>
    y1 = y0 /\ n0 * theta0 = n0 * theta1) /\
  
   (is_valid_ray_at_interface 
   (y0,theta0) (y1,theta1) n0 n1 (spherical_t R) <=> 
      let phi_i = theta0 + y1 / R and phi_t = theta1 + y1 / R in
      y1 = y0 /\ n0 * phi_i = n1 * phi_t) /\
  
   (is_valid_ray_at_interface 
   (y0,theta0) (y1,theta1) n0 n1 (spherical_r R) <=> 
      let phi = y1 / R - theta0  in
      y1 = y0 /\ theta1 = --(theta0 + &2 * phi)) /\   
  
   (is_valid_ray_at_interface 
   (y0,theta0) (y1,theta1) n0 n1 (pcm) <=> 
      y1 = y0 /\ theta1 = --theta0)   /\

   (is_valid_ray_at_interface 
   (y0,theta0) (y1,theta1) n0 n1 (unknown a b c d) <=>
    Cx y1 =  a * Cx y0 + b * Cx theta0  /\  Cx theta1 = c * Cx y0 + d * Cx theta0)`;;


let last_single_ray = define
  `last_single_ray ((sr1,sr2,[]):ray) = sr2 /\
  last_single_ray ((sr1,sr2,CONS sr3 t):ray) = SND (LAST (CONS sr3 t))`;;

let fst_single_ray = define
  `fst_single_ray ((sr1,sr2,rs):ray )= sr1`;; 



let is_valid_ray_in_system = define `
         ( is_valid_ray_in_system (sr1,sr2,[]) (CONS h cs,fs) <=> F) /\
         ( is_valid_ray_in_system (sr1,sr2,CONS h' rs) ([],fs) <=> F) /\
         ( is_valid_ray_in_system ((y0,theta0),(y1,theta1),[]) ([],n,d) <=>
          is_valid_ray_in_free_space (y0,theta0) (y1,theta1) (n,d)) /\
         ( is_valid_ray_in_system ((y0,theta0),(y1,theta1),CONS ((y2,theta2),y3,theta3) rs)
          (CONS ((n',d'),i) cs,n,d) <=>
          (is_valid_ray_in_free_space (y0,theta0) (y1,theta1) (n',d') /\
           is_valid_ray_at_interface (y1,theta1) (y2,theta2) n'
           (head_index (cs,n,d))
           i) /\
           is_valid_ray_in_system ((y2,theta2),(y3,theta3),rs) (cs,n,d))` ;;


(* ------------------------------------------------------------------------- *)
(* Matrix formalism                                                          *)
(* ------------------------------------------------------------------------- *)

let free_space_matrix = new_definition
  `free_space_matrix (n,d) = cmat2x2 (Cx(&1)) (Cx(d)) (Cx(&0)) (Cx(&1))`;;


let interface_matrix = define 
  `interface_matrix n0 n1 plane_t = cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(&0)) (Cx(n0/n1)) /\
   interface_matrix n0 n1 plane_r = cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(&0)) (Cx(&1))    /\
   interface_matrix n0 n1 (spherical_t R)  
    = cmat2x2 (Cx(&1)) (Cx(&0)) (Cx((n0-n1)/n1 * &1/R)) (Cx(n0/n1))                /\
   interface_matrix n0 n1 (spherical_r R)
    = cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(-- &2/R)) (Cx(&1))                             /\
   interface_matrix n0 n1 pcm = cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(&0)) (Cx(-- &1))     /\
   interface_matrix n0 n1 (unknown a b c d) = cmat2x2 (a) (b) (c) (d) `;;



let common_prove t = prove (t,
   SIMP_TAC[free_space_matrix;FORALL_FREE_SPACE_THM;CMAT2X2_VECTOR_MUL;CVECTOR2_EQ;
   LET_END_DEF;is_valid_ray_in_free_space;COMPLEX_MUL_LZERO;COMPLEX_MUL_LID;LET_DEF;
   COMPLEX_ADD_LID;interface_matrix;is_valid_ray_at_interface;COMPLEX_ADD_RID;RE;
   GSYM CX_ADD;GSYM CX_MUL;GSYM CX_DIV;CX_INJ]
   THEN CONV_TAC REAL_FIELD);;
 
let FS_MATRIX = common_prove
  `!fs y0 theta0 y1 theta1.
    is_valid_free_space fs
    /\ is_valid_ray_in_free_space (y0,theta0) (y1,theta1) fs
    ==> vector[Cx(y1);Cx(theta1)] = free_space_matrix fs ** vector[Cx(y0);Cx(theta0)]`;;


let PI_MATRIX_TRANSMITTED = common_prove
  `!n0 n1 y0 theta0 y1 theta1.
    is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 plane_t
    /\ &0 < n0 /\ &0 < n1
      ==> vector [Cx y1;Cx theta1]
        = interface_matrix n0 n1 plane_t ** vector [Cx y0; Cx theta0]`;;

let PI_MATRIX_REFLECTED = common_prove
  `!n0 y0 theta0 y1 theta1.
    is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 plane_r
    /\ &0 < n0 /\ &0 < n1
      ==> vector [Cx y1;Cx theta1]
        = interface_matrix n0 n1 plane_r ** vector [Cx y0;Cx theta0]`;;



let SI_MATRIX_TRANSMITTED = common_prove
  `!n0 n1 y0 theta0 y1 theta1 R.
   is_valid_interface (spherical R) /\ &0 < n0 /\  &0 < n1
   /\ is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 (spherical_t R)
     ==> vector [Cx y1;Cx theta1]
      = interface_matrix n0 n1 (spherical_t R)  ** vector [Cx y0;Cx theta0]`;;

let SI_MATRIX_REFLECTED = common_prove
  `!n0 y0 theta0 y1 theta1 R.
   is_valid_interface (spherical R) /\ &0 < n0
   /\ is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 (spherical_r R)
     ==> vector [Cx y1;Cx theta1]
      = interface_matrix n0 n1 (spherical_r R)  ** vector [Cx y0; Cx theta0]`;;


let PCM_MATRIX = common_prove
  `!n0 y0 theta0 y1 theta1.
    is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 pcm
    /\ &0 < n0 /\ &0 < n1
      ==> vector [Cx y1;Cx theta1]
        = interface_matrix n0 n1 pcm ** vector [Cx y0;Cx theta0]`;;


let UNKNOWN_MATRIX = common_prove
  `!n0 y0 theta0 y1 theta1 a b c d.
    is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 (unknown a b c d)
    /\ &0 < n0 /\ &0 < n1
      ==> vector [Cx y1;Cx theta1]
        = interface_matrix n0 n1 (unknown a b c d) ** vector [Cx y0;Cx theta0]`;;

(*-------General Relationship for any Interface i --------------------*)

let INTERFACE_MATRIX = prove
  (`!n0 n1 y0 theta0 y1 theta1 i.
  is_valid_interface i /\ &0 < n0 /\ &0 < n1 /\
  is_valid_ray_at_interface (y0,theta0) (y1,theta1) n0 n1 i 
    ==> vector [Cx y1;Cx theta1] = interface_matrix n0 n1 i  ** vector [Cx y0; Cx theta0]`,
  REPEAT (MATCH_MP_TAC interface_IND ORELSE GEN_TAC)
  THEN MESON_TAC[PI_MATRIX_TRANSMITTED;PI_MATRIX_REFLECTED;SI_MATRIX_TRANSMITTED;
                 SI_MATRIX_REFLECTED;PCM_MATRIX;UNKNOWN_MATRIX]);;


let system_composition = define 
  `system_composition ([],fs) = free_space_matrix fs /\ 
   system_composition (CONS ((n',d'),i) cs,n,d) =
     system_composition (cs,n,d)
     ** interface_matrix n' (head_index (cs,n,d)) i
     ** free_space_matrix (n',d')`;;


let IS_VALID_OPTICAL_SYSTEM_REC = prove
  (`!fs fs' i t.
    is_valid_optical_system ([],fs) = is_valid_free_space fs /\
    (is_valid_optical_system (CONS (fs,i) t,fs') <=> is_valid_free_space fs
    /\ is_valid_interface i /\ is_valid_optical_system (t,fs'))`,
  REWRITE_TAC[is_valid_optical_system;ALL;is_valid_optical_component]
  THEN MESON_TAC[]);;

let LAST_SINGLE_RAY_REC = prove
  (`!sr1 sr2 sr3 sr4 t.
    last_single_ray (sr1,sr2,[]) = sr2 /\
    last_single_ray (sr1,sr2,CONS (sr3,sr4) t) = last_single_ray (sr3,sr4,t)`,
  REPEAT (LIST_INDUCT_TAC ORELSE GEN_TAC)
  THEN SIMP_TAC[last_single_ray;LAST;NOT_CONS_NIL]);;

let IS_VALID_OPTICAL_SYSTEM_HEAD_INDEX = prove
  (`!os:optical_system. is_valid_optical_system os ==> &0 < head_index os`,
  REWRITE_TAC[FORALL_OPTICAL_SYSTEM_THM] THEN MATCH_MP_TAC list_INDUCT
  THEN SIMP_TAC[FORALL_FREE_SPACE_THM;IS_VALID_OPTICAL_SYSTEM_REC;ALL;
    is_valid_free_space;head_index;FORALL_OPTICAL_COMPONENT_THM]);;


let wrap f x = f [x];;

let SYSTEM_MATRIX = prove
  (`!sys ray.
    is_valid_optical_system sys /\ is_valid_ray_in_system ray sys ==> 
      let ((y0,theta0),(y1,theta1),rs) = ray in
      let (y_n,theta_n) = last_single_ray ray in
        vector [Cx(y_n);Cx(theta_n)] = system_composition sys ** vector [Cx(y0);Cx(theta0)]`,
  REWRITE_TAC[FORALL_OPTICAL_SYSTEM_THM;FORALL_RAY_THM]
  THEN MATCH_MP_TAC list_INDUCT THEN CONJ_TAC
  THEN REWRITE_TAC[FORALL_SINGLE_RAY_THM;FORALL_OPTICAL_COMPONENT_THM] 
  THENL [ALL_TAC;REPEAT GEN_TAC THEN DISCH_TAC]
  THEN REPEAT (MATCH_MP_TAC list_INDUCT ORELSE GEN_TAC) THEN CONJ_TAC
  THEN REWRITE_TAC[IS_VALID_OPTICAL_SYSTEM_REC;is_valid_ray_in_system;ALL;
    system_composition;head_index]
  THENL [
    REWRITE_TAC[last_single_ray;LET_DEF;LET_END_DEF;FS_MATRIX];
    let lemma = prove(`!P. (!a. P a) <=> !y1 th1 y2 th2. P((y1,th1),y2,th2)`,
      REWRITE_TAC[FORALL_PAIR_THM]) in
    REWRITE_TAC[lemma;is_valid_ray_in_system;GSYM CMATRIX_VECTOR_MUL_ASSOC;
      LAST_SINGLE_RAY_REC;is_valid_free_space]
    THEN REPEAT STRIP_TAC THEN REPEAT LET_TAC
    THEN ASSUM_LIST (REWRITE_TAC o map (REWRITE_RULE[PAIR_EQ] o GSYM))
    THEN REPEAT_GTCL IMP_RES_THEN (wrap REWRITE_TAC)
      (REWRITE_RULE[IMP_CONJ;is_valid_free_space;FORALL_FREE_SPACE_THM]
        (GSYM FS_MATRIX))
    THEN IMP_RES_THEN ASSUME_TAC IS_VALID_OPTICAL_SYSTEM_HEAD_INDEX
    THEN REPEAT_GTCL IMP_RES_THEN (wrap REWRITE_TAC)
      (REWRITE_RULE[IMP_CONJ] (GSYM INTERFACE_MATRIX))
    THEN EVERY_ASSUM (REPEAT_GTCL IMP_RES_THEN MP_TAC o REWRITE_RULE[IMP_CONJ])
    THEN ASM_REWRITE_TAC[LET_DEF;LET_END_DEF] THEN SIMP_TAC[]
  ]);;


let RAY_ABCD_LAW = prove 
 (`!ray say. ~(SND(fst_single_ray ray) = &0) /\ is_valid_optical_system sys /\ 
   is_valid_ray_in_system ray sys ==> 
   let M = system_composition sys in 
   let Q_n = Cx(FST(last_single_ray ray) / SND (last_single_ray ray)) and  
   Q_0 =  Cx(FST(fst_single_ray ray)/ SND(fst_single_ray ray)) in  
   Q_n =  ((M$1$1) * Q_0 + (M$1$2))/((M$2$1) * Q_0 + (M$2$2))`,

 REPEAT STRIP_TAC THEN 
 SUBGOAL_THEN `let ((y0,theta0),(y1,theta1),rs) = ray in
      let (y_n,theta_n) = last_single_ray ray in
        vector [Cx(y_n);Cx(theta_n)] = system_composition sys ** vector [Cx(y0);Cx(theta0)] ` ASSUME_TAC
THENL[ASM_SIMP_TAC[SYSTEM_MATRIX];ALL_TAC] THEN
REPEAT (POP_ASSUM MP_TAC) THEN LET_TAC THEN LET_TAC THEN LET_TAC THEN LET_TAC THEN 
REPEAT(POP_ASSUM MP_TAC) THEN REWRITE_TAC[FST;SND;fst_single_ray] THEN REPEAT STRIP_TAC THEN 
POP_ASSUM MP_TAC THEN 
SIMP_TAC[CMAT2X2_VECTOR_MUL;CVECTOR2_EQ;CMAT2X2_VECTOR_MUL_ALT2] THEN 
REPLICATE_TAC 7 (POP_ASSUM MP_TAC) THEN 
REWRITE_TAC[EQ_SYM_EQ] THEN 
REPEAT STRIP_TAC THEN ONCE_ASM_REWRITE_TAC[] THEN 
ONCE_ASM_REWRITE_TAC[] THEN 
ASM_SIMP_TAC[GSYM CX_INJ; CX_DIV;CX_ADD;CX_MUL;COMPLEX_FIELD 
`!theta0 y0 a b c d. ~(Cx (theta0) = Cx (&0)) ==> (a * Cx(y0) / Cx(theta0) + b) / (c * Cx(y0) / Cx(theta0) 
   + d) = (Cx(theta0)*(a * Cx(y0) / Cx(theta0) + b)) / (Cx(theta0)*(c * Cx(y0) / Cx(theta0) + d))`] THEN 
REPEAT (POP_ASSUM MP_TAC) THEN ONCE_REWRITE_TAC[GSYM CX_INJ] THEN 
SIMP_TAC[GSYM CX_INJ; CX_DIV;CX_ADD;CX_MUL;COMPLEX_FIELD 
`!theta0 a b c d y0. ~( Cx(theta0) = Cx(&0)) ==>  
(a * Cx y0 / Cx theta0 + b) = Cx(&1)/Cx(theta0) * (a * Cx y0  + b* Cx theta0)`;
COMPLEX_FIELD `!a b c . ~(a = Cx(&0)) ==> ((Cx (&1)/ a * b )/ (Cx (&1)/ a * c)  = b/c)`]);;
