(* ========================================================================= *)
(*                                                                           *)
(*                        Q-Parameters and Gaussian Beam                     *)
(*                                                                           *)
(*        (c) Copyright, Umair Siddique, 2012-2015                           *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last updated: February 20,2015                                            *)
(*                                                                           *)
(* ========================================================================= *)


needs "c_ray_optics.ml";;


(* ------------------------------------------------------------------------- *)
(* Q-Parameters                                                              *)
(* ------------------------------------------------------------------------- *)


let rayleigh_range = new_definition
  ` rayleigh_range w0 lam = (pi* (w0 pow 2))/ lam  `;;

let RR = new_definition
  `RR z w0 lam = z*((&1) +  ((rayleigh_range w0 lam)/z) pow 2)`;;

let WW = new_definition
   `WW z w0 lam = w0*(&1 + (z/(rayleigh_range w0 lam)) pow 2) rpow (&1/ &2) `;;

let WWW = new_definition
   `WWW z w0 lam = w0* sqrt(&1 + (z/(rayleigh_range w0 lam)) pow 2) `;;

let WWW_ALT = prove (` !w0 lam. &0 < w0 ==> 
                        sqrt(&1 + (z/(rayleigh_range w0 lam)) pow 2) = 
                        WWW z w0 lam / w0`, 
REWRITE_TAC[WWW] THEN 
IMP_REWRITE_TAC[REAL_EQ_RDIV_EQ] THEN 
SIMP_TAC[REAL_MUL_SYM]);;


let beam_divergence = new_definition `
                      beam_divergence lam w0 = lam / pi * w0`;; 


let qq = new_definition ` 
    qq (z,w0,lam) = (Cx z + ii * Cx (rayleigh_range w0 lam))`;;

let QQ = new_definition
  `QQ z w0 lam =  Cx(&1) /  
   (Cx(&1) / Cx (RR z w0 lam) - ii* Cx(lam)/ (Cx(pi)* Cx((WW z w0 lam) pow 2)))`;;

let COMPLEX_EQ_ZERO =  prove (`
 !a b. ~(a = &0)  /\ ~(b = &0) ==> ~(Cx (a) + ii * (Cx b) = Cx (&0))`,
REWRITE_TAC[GSYM COMPLEX_TRAD] THEN 
SIMP_TAC[COMPLEX_EQ;RE;IM;RE_CX]);;


let COMPLEX_INV_NZ = 
    COMPLEX_FIELD ` !x.  ~(x = Cx(&0)) ==> ~(inv x = Cx(&0))`;;

let RR_ALTERNATIVE_FORM = prove (`
 RR z w0 lam = (z pow 2 + (rayleigh_range w0 lam) pow 2) / z `,
REWRITE_TAC[RR;rayleigh_range] THEN 
ASM_CASES_TAC `z = &0` THEN 
ASM_REWRITE_TAC[] THEN 
REWRITE_TAC[real_div; REAL_MUL_RZERO;REAL_MUL_LZERO; REAL_INV_0] THEN 
REWRITE_TAC[REAL_POW_2] THEN 
SUBGOAL_THEN ` &0 < pi ` ASSUME_TAC THENL[
REWRITE_TAC[PI_POS]; ALL_TAC] THEN 
REPEAT(POP_ASSUM MP_TAC) THEN 
CONV_TAC REAL_FIELD);;


let HALFPOWER_SQUARE_LEMMA = prove (` 
   !(x:real). (&0 < x) ==> (x rpow (&1 / &2)) pow 2 = x`,
   ASM_REWRITE_TAC[] THEN SIMP_TAC[rpow;REAL_POW_2; REAL_FIELD `~(&0 < &0)`;
   REAL_FIELD `~(&1 / &2 = &0 )`;REAL_MUL_RZERO] THEN 
   ASM_SIMP_TAC[rpow;REAL_POW_2;GSYM REAL_EXP_ADD;GSYM REAL_ADD_RDISTRIB;
   REAL_FIELD ` (&1 / &2 + &1 / &2) = &1`; REAL_MUL_LID; EXP_LOG]);;


let SQUARE_POS = prove (` !x. ~(x = &0) ==> &0 < x pow 2 `,
		 ASM_SIMP_TAC[REAL_POW_2; REAL_LT_SQUARE]);;


let WW_POS = prove (
 ` !lam z w0. &0 < w0 /\ &0 < lam /\  &0 <= z ==> &0 < WW z w0 lam`,

REPEAT GEN_TAC THEN REWRITE_TAC [WW;rayleigh_range] THEN
SUBGOAL_THEN ` &0 < pi ` ASSUME_TAC THENL[
REWRITE_TAC[PI_POS]; ALL_TAC] THEN 
REPEAT STRIP_TAC THEN 
ASM_CASES_TAC `z = &0` THENL[ 
ASM_REWRITE_TAC[real_div; REAL_INV_0; REAL_MUL_LZERO;REAL_POW_2;
  REAL_FIELD `&0 < &1`;rpow;REAL_ADD_RID] THEN 
MATCH_MP_TAC REAL_LT_MUL THEN 
ASM_SIMP_TAC[REAL_EXP_POS_LT];ALL_TAC] THEN 
 SUBGOAL_THEN `  &0 < (z / ((pi * w0 pow 2) / lam)) ` ASSUME_TAC THENL[
 MATCH_MP_TAC REAL_LT_DIV THEN
 ASM_SIMP_TAC[REAL_FIELD `&0 <= z /\ ~(z = &0) ==> &0 < z`] THEN
 MATCH_MP_TAC REAL_LT_DIV THEN 
 ASM_SIMP_TAC[] THEN 
 MATCH_MP_TAC REAL_LT_MUL THEN 
 ASM_SIMP_TAC[] THEN 
 MATCH_MP_TAC SQUARE_POS THEN 
 ASM_SIMP_TAC[REAL_POS_NZ ];ALL_TAC] THEN 
 ASM_SIMP_TAC[REAL_FIELD `&0 < x /\ &0 < y ==> &0 < x + y `;REAL_FIELD `&0 < &1`; REAL_POS_NZ; SQUARE_POS;rpow]  THEN
 ASM_SIMP_TAC[rpow;REAL_EXP_POS_LT;REAL_LT_MUL]);;


let  Q_NZ_LEMMA = prove (
   `!w0 lam z . (&0 < w0) /\ &0 < lam  ==> 
   ~(Cx(z) + ii* Cx(rayleigh_range w0 lam) = Cx(&0))`,
REPEAT GEN_TAC THEN SUBGOAL_THEN ` &0 < pi ` ASSUME_TAC THENL[
REWRITE_TAC[PI_POS]; ALL_TAC] THEN POP_ASSUM MP_TAC THEN 
REWRITE_TAC[GSYM COMPLEX_TRAD;COMPLEX_EQ;RE;IM;RE_CX;
           IM_CX;rayleigh_range;DE_MORGAN_THM] THEN 
CONV_TAC REAL_FIELD);;

let Q_NZ = prove (
`!w0 lam z . (&0 < w0) /\ &0 < lam ==> ~(qq (z,w0,lam) = Cx(&0))`,

MESON_TAC[qq;Q_NZ_LEMMA]);;


let RAYLEIGH_POS = prove (`!lam w0.  &0 < w0 /\ &0 < lam ==> 
                   &0 < rayleigh_range w0 lam pow 2 `,
REWRITE_TAC[rayleigh_range] THEN 
REPEAT STRIP_TAC THEN 
MATCH_MP_TAC SQUARE_POS THEN 
MATCH_MP_TAC REAL_POS_NZ THEN 
MATCH_MP_TAC REAL_LT_DIV THEN 
ASM_REWRITE_TAC[PI_POS] THEN 
MATCH_MP_TAC REAL_LT_MUL THEN 
ASM_REWRITE_TAC[REAL_POW_2;PI_POS] THEN 
MATCH_MP_TAC REAL_LT_MUL THEN 
ASM_REWRITE_TAC[]);;


let RAYLEIGH_POS_2 = prove (`!lam w0.  &0 < w0 /\ &0 < lam ==> 
                   &0 < rayleigh_range w0 lam  `,
REWRITE_TAC[rayleigh_range] THEN 
REPEAT STRIP_TAC THEN
MATCH_MP_TAC REAL_LT_DIV THEN 
ASM_REWRITE_TAC[PI_POS] THEN 
MATCH_MP_TAC REAL_LT_MUL THEN 
ASM_REWRITE_TAC[REAL_POW_2;PI_POS] THEN 
MATCH_MP_TAC REAL_LT_MUL THEN 
ASM_REWRITE_TAC[]);;




let ONE_PLUS_X_POS = 
              REAL_FIELD ` !x. &0 < x ==> &0 < &1 + x `;;
let A_PLUS_INV_B = REAL_FIELD `
                   !a b c. ~(c = &0) ==> a + b / c = (a * c + b) / c`;;

let SIMP_LEMMA = 
         REAL_FIELD `!x z. 
         ~(x = &0)  /\ ~(z = &0) ==>  inv z * z pow 2 / x  = z / x`;;


let RAYLEIGH_RANGE_NZ = prove (`
    !w0 lam. &0 < w0 /\ &0 < lam ==>
             ~(rayleigh_range w0 lam pow 2 = &0)`,
MP_TAC PI_POS THEN 
REWRITE_TAC[rayleigh_range] THEN CONV_TAC REAL_FIELD );;

let RAYLEIGH_RANGE_NZ_2 = prove (`
    !w0 lam. &0 < w0 /\ &0 < lam ==>
             ~(rayleigh_range w0 lam = &0)`,
MP_TAC PI_POS THEN 
REWRITE_TAC[rayleigh_range] THEN CONV_TAC REAL_FIELD );;

let COMPLEX_INV_2 = 
    COMPLEX_FIELD `!x y.  ~(x = Cx(&0))  /\  ~(y = Cx(&0)) ==> 
                     ((inv x = inv y) = (x = y))`;;
let COMPLEX_INV_EQ = 
    COMPLEX_FIELD ` !x y.  ~(x = Cx(&0))  /\  ~(y = Cx(&0)) ==> 
                     ( (x = y) = (inv x = inv y))`;;

let Q_PARAMETERS_ALTERNATIVE_FORM = prove
 (`! w0 lam (z:real).  &0 < w0 /\ &0 < lam /\ ~(z = &0) ==> 
  Cx (&1)/ QQ z w0 lam = Cx(&1)/ (Cx(z) + ii* Cx(rayleigh_range w0 lam))`,
  
 REPEAT GEN_TAC THEN REWRITE_TAC [GSYM COMPLEX_TRAD] THEN
  SUBGOAL_THEN
    `!z w0 lam. Cx (&1) / complex (z,rayleigh_range w0 lam) = 
                inv (complex (z,rayleigh_range w0 lam))`
    ASSUME_TAC THENL
   [SIMP_TAC [complex_div; COMPLEX_MUL_LID]; ALL_TAC] THEN
  ASM_REWRITE_TAC [] THEN REWRITE_TAC [COMPLEX_INV_CNJ] THEN
  REWRITE_TAC [cnj; complex_norm; RE; IM] THEN REWRITE_TAC [COMPLEX_TRAD] THEN
  SIMP_TAC [COMPLEX_FIELD `(a + ii*b)/c = a/c + ii*b/c`; CX_DIV] THEN
  SUBGOAL_THEN `&0 < pi` ASSUME_TAC THENL [REWRITE_TAC [PI_POS]; ALL_TAC] THEN
  REPEAT STRIP_TAC THEN
  SUBGOAL_THEN
    `Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2)) pow 2 = 
                Cx (z pow 2 + rayleigh_range w0 lam pow 2)`
    ASSUME_TAC THENL
   [REWRITE_TAC [CX_INJ; COMPLEX_POW_2; GSYM CX_MUL] THEN
    REWRITE_TAC [GSYM REAL_POW_2] THEN IMP_REWRITE_TAC [SQRT_POW_2] THEN
    MATCH_MP_TAC REAL_LT_IMP_LE THEN MATCH_MP_TAC REAL_LT_ADD THEN
    ASM_SIMP_TAC [SQUARE_POS; RAYLEIGH_POS];
    ALL_TAC] THEN
  ONCE_ASM_REWRITE_TAC [] THEN REWRITE_TAC [GSYM CX_DIV] THEN
  REWRITE_TAC
    [QQ; complex_div; COMPLEX_MUL_LID; COMPLEX_INV_INV; GSYM CX_INV;
     GSYM CX_MUL] THEN
  SIMP_TAC [COMPLEX_FIELD `!a b. a - ii * b = a + ii *(--b)`] THEN
  REWRITE_TAC [GSYM CX_NEG] THEN REWRITE_TAC [GSYM COMPLEX_TRAD] THEN
  REWRITE_TAC [COMPLEX_EQ] THEN REWRITE_TAC [RE; IM] THEN CONJ_TAC THENL
   [REWRITE_TAC [RR; REAL_INV_MUL] THEN REWRITE_TAC [REAL_POW_DIV] THEN
    IMP_REWRITE_TAC [A_PLUS_INV_B] THEN IMP_REWRITE_TAC [REAL_POS_NZ] THEN
    ASM_SIMP_TAC [SQUARE_POS] THEN
    REWRITE_TAC [REAL_MUL_LID; real_div; REAL_INV_INV; REAL_INV_MUL] THEN
    REWRITE_TAC [REAL_FIELD `(a:real)*b*c = (a*c)*b`] THEN
    REWRITE_TAC [GSYM real_div; REAL_POW_2] THEN
    REWRITE_TAC [REAL_FIELD `(a:real)*b*c = (a*c)*b`] THEN
    IMP_REWRITE_TAC [REAL_MUL_LINV] THEN SIMP_TAC [REAL_MUL_LID];
    SIMP_TAC [REAL_FIELD `! y a x. (-- (x:real) = --y/ a) = (x = y / a)`] THEN
    SIMP_TAC [REAL_INV_MUL; WW] THEN SIMP_TAC [REAL_POW_MUL] THEN
    SIMP_TAC [REAL_POW_MUL] THEN IMP_REWRITE_TAC [HALFPOWER_SQUARE_LEMMA] THEN
    CONJ_TAC THENL
     [ALL_TAC;
      SIMP_TAC [REAL_POW_DIV; RAYLEIGH_RANGE_NZ; SQUARE_POS; REAL_POS_NZ] THEN
      IMP_REWRITE_TAC [REAL_FIELD `&0 < x ==> &0 < &1 + x`] THEN
      IMP_REWRITE_TAC [REAL_LT_DIV] THEN
      ASM_SIMP_TAC
        [REAL_POW_DIV; RAYLEIGH_RANGE_NZ; SQUARE_POS; REAL_POS_NZ;
         RAYLEIGH_POS]] THEN
    SIMP_TAC [REAL_INV_MUL] THEN REWRITE_TAC [REAL_POW_DIV] THEN
    IMP_REWRITE_TAC [A_PLUS_INV_B] THEN ASM_SIMP_TAC [RAYLEIGH_RANGE_NZ] THEN
    REWRITE_TAC [REAL_MUL_LID] THEN
    SIMP_TAC [real_div; REAL_INV_MUL; REAL_INV_INV] THEN
    SIMP_TAC
      [REAL_FIELD
         `(lam *inv pi * inv (w0 pow 2) *
 inv (rayleigh_range w0 lam pow 2 + z pow 2) *
 rayleigh_range w0 lam pow 2) = 
 (lam *(inv pi * inv (w0 pow 2)) * (rayleigh_range w0 lam pow 2 
 *inv (rayleigh_range w0 lam pow 2 + z pow 2)))`] THEN
    REWRITE_TAC [GSYM REAL_INV_MUL; GSYM real_div] THEN
    SIMP_TAC [REAL_FIELD `(a:real)*b*c = (a*b)*c`] THEN
    REWRITE_TAC [GSYM real_div] THEN
    ASSERT_TAC `lam / (pi * w0 pow 2) = inv (rayleigh_range w0 lam)` THENL
     [SIMP_TAC [rayleigh_range; real_div; REAL_INV_INV] THEN
      REWRITE_TAC [REAL_INV_MUL; REAL_INV_INV] THEN REAL_ARITH_TAC;
      ASM_REWRITE_TAC [] THEN REWRITE_TAC [real_div; REAL_POW_2] THEN
      SIMP_TAC [REAL_FIELD `(a:real)*(c*c)*d = (c*a)*(c*d)`] THEN
      IMP_REWRITE_TAC [REAL_MUL_RINV] THEN
      ASM_SIMP_TAC [REAL_MUL_LID; RAYLEIGH_RANGE_NZ_2] THEN
      CONV_TAC REAL_FIELD]]);;

 let qq_QQ = prove (` !w0 lam z.
         &0 < w0 /\ &0 < lam /\ ~(z = &0) ==>
    Cx(&1) / qq (z,w0,lam) = Cx(&1) / QQ z w0 lam`,
REPEAT STRIP_TAC THEN 
IMP_REWRITE_TAC[Q_PARAMETERS_ALTERNATIVE_FORM] THEN 
MESON_TAC[qq]);;

let qq_QQ_EXPLICIT = prove (`!w0 lam z.
         &0 < w0 /\ &0 < lam /\ ~(z = &0) ==>
 inv (qq (z,w0,lam)) = (Cx (&1) / Cx (RR z w0 lam) -
                       ii * Cx lam / (Cx pi * Cx (WW z w0 lam pow 2)))`,
REPEAT STRIP_TAC THEN 
MP_TAC (ISPECL[`w0:real`;`lam:real`;`z:real`] qq_QQ) THEN 
ASM_SIMP_TAC[] THEN REWRITE_TAC[complex_div;COMPLEX_MUL_LID;QQ] THEN 
SIMP_TAC[COMPLEX_INV_INV]);;



let REAL_INV_NZ = 
    REAL_FIELD ` !x. ~(x = &0) ==> ~(inv x = &0)`;;

let REAL_INV_NZ_NEG = 
    REAL_FIELD ` !x. ~(x = &0)  /\ ~(z = &0)  ==> ~(--(z * inv x) = &0)`;;

let REAL_NZ_POW_2 = 
    REAL_FIELD ` !x. ~(x = &0)   ==> ~(x pow 2 = &0)`;;


let RR_NZ = prove (
    ` !w0 lam z. ~(z = &0) /\ (rayleigh_range w0 lam pow 2 = (z pow 2)) ==>
                               ~(RR z w0 lam = &0)`,
REPEAT GEN_TAC THEN 
DISCH_TAC THEN 
SIMP_TAC[RR;REAL_ENTIRE;DE_MORGAN_THM;PI_NZ] THEN
ASM_SIMP_TAC[] THEN
IMP_REWRITE_TAC[REAL_FIELD `~((rayleigh_range w0 lam / z) pow 2 = -- &1)
                   ==> ~(&1 + (rayleigh_range w0 lam / z) pow 2 = &0)`] THEN
SIMP_TAC[REAL_POW_DIV] THEN
SIMP_TAC[REAL_NZ_POW_2;
         REAL_FIELD ` !x y. ~(y = &0) ==> --x / y = -- (x / y)`] THEN 
ASM_SIMP_TAC[] THEN 
ASM_SIMP_TAC[REAL_DIV_REFL;REAL_NZ_POW_2] THEN 
REAL_ARITH_TAC);;

let RR_NZ_ABS = prove (
    ` !w0 lam z. ~(z = &0) /\ ( abs (rayleigh_range w0 lam) = (abs z)) ==>
                               ~(RR z w0 lam = &0)`,
MESON_TAC[RR_NZ;REAL_EQ_SQUARE_ABS]);;


let QQ_NZ = prove (
  ` !w0 lam z.  &0 < z  /\ ( abs (rayleigh_range w0 lam) = (abs z)) /\ 
                &0 < w0 /\ &0 < lam ==> ~(QQ z w0 lam  = Cx(&0))`,

REWRITE_TAC[QQ;complex_div;COMPLEX_MUL_LID] THEN 
IMP_REWRITE_TAC[COMPLEX_INV_NZ] THEN 
SIMP_TAC[COMPLEX_FIELD `!a b. a - ii * b = a + ii *(--b)` ] THEN
SIMP_TAC[GSYM CX_NEG;GSYM CX_MUL;GSYM CX_INV] THEN 
IMP_REWRITE_TAC[COMPLEX_EQ_ZERO] THEN 
IMP_REWRITE_TAC[REAL_INV_NZ] THEN 
IMP_REWRITE_TAC[REAL_INV_NZ_NEG] THEN 
SIMP_TAC[REAL_ENTIRE;DE_MORGAN_THM;PI_NZ] THEN
IMP_REWRITE_TAC[REAL_NZ_POW_2] THEN
IMP_REWRITE_TAC[RR_NZ] THEN 
IMP_REWRITE_TAC[REAL_POS_NZ;WW_POS] THEN 
MESON_TAC[RR_NZ;REAL_EQ_SQUARE_ABS;REAL_LE_LT]);;

(*===========================================================================*)
(*              FORMALIZATION OF BEAMS                                       *)
(*===========================================================================*)


new_type_abbrev("sing_beam",`:real#real`);;

let FORALL_SINGLE_BEAM_THM = prove
  (`!P. (!sr:sing_beam. P sr) <=> !z w0. P (z,w0)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

let FORALL_SINGLE_BEAM_PAIR_THM = prove
  (`!P. (!sr:sing_beam#sing_beam. P sr) <=> !(a:sing_beam) b. P (a,b)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;


new_type_abbrev("beam",
  `:sing_beam # sing_beam # (sing_beam # sing_beam) list`);;

let FORALL_BEAM_THM = prove
  (`!P. (!r:beam. P r) <=> !sr1 sr2 srs. P (sr1,sr2,srs)`,
  REWRITE_TAC[FORALL_PAIR_THM]);;

let is_valid_single_beam = define `
   is_valid_single_beam ((z,w0):sing_beam) <=>
   &0 < w0 /\ ~(z = &0)`;;



let is_valid_beam_in_free_space = define 
  `is_valid_beam_in_free_space (z,w0) (z',w0') ((n0,d):free_space) <=>
   w0' = w0  /\ z' = z + d`;;


let is_valid_beam_at_plane_inerf = define 
  `(is_valid_beam_at_plane_inerf (z,w0) (z',w0') n0 n1
                                   plane transmitted  <=>
    z' = n1/ n0 * z  /\ w0' =  (sqrt (n1/ n0)) * w0)  /\ 
   (is_valid_beam_at_plane_inerf (z,w0) (z',w0') n0 n1
                                   plane reflected  <=>
    z' =  z  /\ w0' =  w0)`;;
 
let is_valid_beam_at_inerface = define `
    (is_valid_beam_at_inerface (z,w0) (z',w0') lam n n' plane transmitted <=>
    (z' = n' / n * z)  /\ (w0' =  (sqrt (n' / n)) * w0) /\  &0 < n /\ &0 < n')

/\ (is_valid_beam_at_inerface (z,w0) (z',w0') lam n n' plane reflected <=>
    z' =  z  /\ w0' =  w0 /\  &0 < n  /\ &0 < n')

/\ (is_valid_beam_at_inerface (z,w0) (z',w0') lam n n' (spherical R) transmitted <=>
    is_valid_single_beam (z,w0) /\ is_valid_single_beam (z',w0')/\ &0 < n /\ &0 < n'/\ 
    ~(Cx (--(n - n') / (n * R)) = Cx (&1) / qq (z,w0,lam)) /\
     (&1 / RR z' w0' lam = n / n' * &1 / RR z w0 lam + (n - n') / (n' * R)) /\
     (WW z' w0' lam = sqrt (n' / n) * WW z w0 lam))

/\ (is_valid_beam_at_inerface (z,w0) (z',w0') lam n n' (spherical R) reflected <=>
    is_valid_single_beam (z,w0) /\ is_valid_single_beam (z',w0')/\ &0 < n /\ &0 < n'/\ 
    ~(qq (z,w0,lam) = Cx (R / &2)) /\
     inv (RR z' w0' lam) = inv (RR z w0 lam) - &2 / R /\
     WW z' w0' lam = WW z w0 lam)`;;



let  is_valid_beam_in_system = define `
     (is_valid_beam_in_system (sr1,sr2,[]) lam (CONS h cs,fs) <=> F) /\
     (is_valid_beam_in_system (sr1,sr2,CONS h' rs) lam ([],fs) <=> F) /\
     (is_valid_beam_in_system ((z,w0),(z',w0'),[]) lam ([],n,d) <=>
      is_valid_beam_in_free_space (z,w0) (z',w0') (n,d)) /\
      (is_valid_beam_in_system
          ((z,w0),(z',w0'),CONS ((z'',w0''),z''',w0''') rs) lam
          (CONS ((n',d'),i,ik) cs,n,d) <=>
          (is_valid_beam_in_free_space (z,w0) (z',w0') (n',d') /\
           is_valid_beam_at_inerface (z',w0') (z'',w0'') lam n'
           (head_index (cs,n,d)) i ik)  /\  
      is_valid_beam_in_system ((z'',w0''),(z''',w0'''),rs) lam (cs,n,d))`;;



let ABCD_FS = prove (` !n d w0 w0' z z' lam A B C D.
   is_valid_beam_in_free_space (z,w0) (z',w0') (n,d) /\ 
   cmat2x2 A B C D = free_space_matrix (n,d) ==> 
   qq (z',w0',lam) =  (A * qq (z,w0,lam) + B)/ (C * qq (z,w0,lam) + D)`,

SIMP_TAC[is_valid_free_space;CMAT2X2_EQ; free_space_matrix] THEN
SIMP_TAC[COMPLEX_MUL_LID;COMPLEX_MUL_LZERO;COMPLEX_ADD_LID;complex_div;
           COMPLEX_MUL_RID;COMPLEX_INV_1; is_valid_beam_in_free_space] THEN
REWRITE_TAC[qq;CX_ADD;GSYM COMPLEX_ADD_ASSOC] THEN
MESON_TAC[COMPLEX_ADD_SYM]);;

let RAYLEIGH_CMUL = prove (` ! w0 lam a. 
              rayleigh_range (a*w0) lam = a pow 2 * rayleigh_range w0 lam`,

SIMP_TAC[rayleigh_range;REAL_POW_MUL] THEN
REAL_ARITH_TAC);;

let INV_COMPLEX_INV = prove (`inv (Cx (n / n')) = Cx (n' / n)`,

SIMP_TAC[CX_DIV;complex_div;COMPLEX_INV_MUL;COMPLEX_INV_INV] THEN
CONV_TAC COMPLEX_FIELD);;



let ABCD_PLANE_INTEFACE = prove (` !ik d w0 w0' z z' n n' lam A B C D.
   is_valid_beam_at_plane_inerf (z,w0) (z',w0') n n' plane ik  /\
   &0 < n  /\ &0 < n' /\
   cmat2x2 A B C D = interface_matrix n n' plane ik  ==> 
   qq (z',w0',lam) =  (A * qq (z,w0,lam) + B)/ (C * qq (z,w0,lam) + D)`,

MATCH_MP_TAC interface_kind_IND THEN 
CONJ_TAC THENL[

SIMP_TAC[ is_valid_beam_at_plane_inerf;interface_matrix;CMAT2X2_EQ] THEN
SIMP_TAC[COMPLEX_MUL_LID;COMPLEX_MUL_LZERO;COMPLEX_ADD_LID;complex_div;
         COMPLEX_MUL_RID;COMPLEX_INV_1; COMPLEX_MUL_LZERO;
         COMPLEX_MUL_RZERO; COMPLEX_ADD_RID]
 THEN
REWRITE_TAC[qq;CX_ADD;COMPLEX_POW_MUL] THEN
SIMP_TAC[RAYLEIGH_CMUL;INV_COMPLEX_INV] THEN
ASM_SIMP_TAC[REAL_LT_DIV;REAL_LE_LT;
         SQRT_POW_2;CX_MUL;COMPLEX_ADD_RDISTRIB] THEN
REWRITE_TAC[GSYM COMPLEX_MUL_ASSOC] THEN 
MESON_TAC[COMPLEX_MUL_SYM];ALL_TAC] THEN 

SIMP_TAC[ is_valid_beam_at_plane_inerf;interface_matrix;CMAT2X2_EQ] THEN
SIMP_TAC[COMPLEX_MUL_LID;COMPLEX_MUL_LZERO;COMPLEX_ADD_LID;complex_div;
         COMPLEX_MUL_RID;COMPLEX_INV_1; COMPLEX_MUL_LZERO;
         COMPLEX_MUL_RZERO; COMPLEX_ADD_RID]);;

let LCM_LEMMA = 
 COMPLEX_FIELD ` !a b c d x. ~(d = Cx(&0)) /\ ~(b = Cx(&0))==>
           ((a:complex)/b) * x + c/d = (a*d*x + c*b)/(b*d)  `;;

let COMPLEX_DIV_NZ = 
COMPLEX_FIELD `!a b. ~(a = Cx(&0)) /\ ~(b = Cx(&0)) ==> ~(a / b = Cx(&0))`;;

let COMPLEX_MUL_ALT = COMPLEX_FIELD `!(a:complex) b c.  ~(b = Cx(&0)) /\ ~(c = Cx(&0))/\ ~(a / c = inv  b) ==>  ~(a*b = c)`;;



let BEAM_SI_TRANS = prove
 (`!R w0 w0' z z' n n' lam A B C D.
  
   &0 < n  /\ &0 < n' /\ ~(R = &0) /\ &0 < w0 /\ &0 < lam /\ &0 < w0' /\
  ~(z' = &0) /\ ~(z = &0) /\
   ~(Cx (--(n - n') / (n * R)) = Cx(&1)/ (qq (z,w0,lam)))/\
   (&1) / (RR z' w0' lam) = (n/n')*((&1) / (RR z w0 lam)) + (n - n')/(n'*R)/\
    WW z' w0' lam = sqrt (n'/n) * WW z w0 lam /\
   cmat2x2 A B C D = interface_matrix n n' (spherical R) transmitted  ==> 
   qq (z',w0',lam) =  (A * qq (z,w0,lam) + B)/ (C * qq (z,w0,lam) + D)`,
  SIMP_TAC [is_valid_beam_at_plane_inerf; interface_matrix; CMAT2X2_EQ] THEN
  SIMP_TAC
    [COMPLEX_MUL_LID; COMPLEX_MUL_LZERO; COMPLEX_ADD_LID; complex_div;
     COMPLEX_MUL_RID; COMPLEX_INV_1; COMPLEX_MUL_LZERO; COMPLEX_MUL_RZERO;
     COMPLEX_ADD_RID] THEN
  REWRITE_TAC [GSYM complex_div] THEN
  ASM_SIMP_TAC
    [REAL_FIELD
       `!R n'. ~(R = &0) /\ ~(n' = &0) ==> ((n - n') / n' * &1 / R) = (n - n') / (n' * R)`;
     REAL_POS_NZ] THEN
  REWRITE_TAC [CX_DIV] THEN IMP_REWRITE_TAC [LCM_LEMMA] THEN
  SIMP_TAC [CX_INJ] THEN
  ASM_SIMP_TAC [REAL_POS_NZ; REAL_ENTIRE; DE_MORGAN_THM] THEN
  REWRITE_TAC [complex_div; COMPLEX_INV_INV; COMPLEX_INV_MUL] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `!(a:complex) b c d. 
                  a*b*c*d = (c*d*a)*b`] THEN
  REWRITE_TAC [GSYM complex_div] THEN REPEAT STRIP_TAC THEN
  REWRITE_TAC [GSYM CX_MUL] THEN
  ASSERT_TAC
    `qq (z',w0',lam) =
 (Cx (n' * R) * Cx n' * qq (z,w0,lam)) /
 (Cx (n - n') * Cx n' * qq (z,w0,lam) + Cx (n * n' * R)) <=>
inv (qq (z',w0',lam)) =
 inv ((Cx (n' * R) * Cx n' * qq (z,w0,lam)) /
 (Cx (n - n') * Cx n' * qq (z,w0,lam) + Cx (n * n' * R)))` THENL
   [MATCH_MP_TAC (GSYM COMPLEX_INV_2) THEN IMP_REWRITE_TAC [GSYM Q_NZ] THEN
    IMP_REWRITE_TAC [GSYM COMPLEX_DIV_NZ] THEN
    SIMP_TAC [COMPLEX_FIELD `!(a:complex). Cx(&0) = b <=> b = Cx(&0)`] THEN
    SIMP_TAC [COMPLEX_ENTIRE; DE_MORGAN_THM] THEN
    ASM_SIMP_TAC [CX_INJ; REAL_ENTIRE; REAL_POS_NZ; DE_MORGAN_THM; Q_NZ] THEN
    SIMP_TAC
      [COMPLEX_FIELD `!(a:complex) b. ~(a + b = Cx(&0)) = ~(a = -- b )`] THEN
    SIMP_TAC [GSYM CX_MUL; COMPLEX_MUL_ASSOC] THEN REWRITE_TAC [GSYM CX_NEG] THEN
    MATCH_MP_TAC COMPLEX_MUL_ALT THEN
    ASM_SIMP_TAC
      [REAL_FIELD `~(--(n * n' * R) = &0) = ~((n * n' * R) = &0)`; Q_NZ;
       CX_INJ; REAL_ENTIRE; DE_MORGAN_THM; REAL_POS_NZ] THEN
    SIMP_TAC [GSYM CX_DIV] THEN
    IMP_REWRITE_TAC
      [REAL_FIELD
         `~(n' = &0)/\ ~(n = &0) /\ ~(R = &0) ==> (((n - n') * n') / --(n * n' * R)) = --(n - n')/(n*R)`] THEN
    ASM_SIMP_TAC [REAL_POS_NZ; COMPLEX_MUL_LID; CX_DIV];
    ASM_REWRITE_TAC [] THEN
    SIMP_TAC [complex_div; COMPLEX_INV_INV; COMPLEX_INV_MUL] THEN
    ONCE_REWRITE_TAC [COMPLEX_MUL_SYM] THEN SIMP_TAC [GSYM COMPLEX_INV_MUL] THEN
    SIMP_TAC [COMPLEX_ADD_RDISTRIB] THEN SIMP_TAC [GSYM complex_div] THEN
    SIMP_TAC [GSYM CX_MUL; GSYM CX_DIV] THEN
    IMP_REWRITE_TAC
      [COMPLEX_FIELD
         `!(a:complex) b c d. 
            ~(b = Cx(&0)) /\  ~(c = Cx(&0)) ==> (a*b*c)/(d*b*c) = a/d`] THEN
    ASM_SIMP_TAC [CX_INJ; REAL_POS_NZ; Q_NZ] THEN SIMP_TAC [CX_MUL] THEN
    IMP_REWRITE_TAC
      [COMPLEX_FIELD
         `!(a:complex) b c d. 
            ~(b = Cx(&0)) /\  ~(c = Cx(&0)) ==>
     ((a*b*c)/((b*c)*b*d)) = (a/(b*d))`] THEN
    ASM_SIMP_TAC [CX_INJ; REAL_POS_NZ; Q_NZ] THEN
    ASSERT_TAC
      `Cx(&1) / qq (z',w0',lam) = (Cx n / Cx n') *Cx(&1) / qq (z,w0,lam) +
                                          Cx(n - n')/ Cx(n' * R)` THENL
     [ALL_TAC;
      POP_ASSUM MP_TAC THEN
      SIMP_TAC
        [complex_div; COMPLEX_INV_INV; COMPLEX_INV_MUL; COMPLEX_MUL_LID] THEN
      DISCH_TAC THEN SIMP_TAC [CX_MUL; GSYM CX_INV; REAL_INV_MUL] THEN
      CONV_TAC COMPLEX_FIELD] THEN
    IMP_REWRITE_TAC [qq_QQ] THEN
    SIMP_TAC [QQ; COMPLEX_INV_INV; complex_div; COMPLEX_MUL_LID] THEN
    SIMP_TAC [COMPLEX_SUB_LDISTRIB] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `(Cx n * inv (Cx n')) * ii * Cx lam * inv (Cx pi * Cx (WW z w0 lam pow 2)) =  ii *  (Cx n * inv (Cx n')) * Cx lam * inv (Cx pi * Cx (WW z w0 lam pow 2))`] THEN
    SIMP_TAC [GSYM CX_MUL; GSYM CX_INV] THEN
    SIMP_TAC [COMPLEX_FIELD `!(a:complex) b c.  a - b + c = (a + c) - b`] THEN
    SIMP_TAC [GSYM CX_INV; GSYM CX_MUL] THEN
    SIMP_TAC [COMPLEX_FIELD `!a b. a - ii * b = a + ii *(--b)`] THEN
    REWRITE_TAC [GSYM CX_NEG] THEN REWRITE_TAC [GSYM COMPLEX_TRAD] THEN
    REWRITE_TAC [COMPLEX_EQ] THEN REWRITE_TAC [GSYM CX_ADD; RE; IM] THEN
    REWRITE_TAC [GSYM COMPLEX_TRAD] THEN SIMP_TAC [IM; RE] THEN
    ASM_SIMP_TAC [] THEN SIMP_TAC [REAL_POW_MUL] THEN
    IMP_REWRITE_TAC [SQRT_POW_2] THEN SIMP_TAC [REAL_LE_LT] THEN
    ASM_SIMP_TAC [REAL_LT_DIV] THEN SIMP_TAC [REAL_INV_MUL] THEN CONJ_TAC THENL
     [UNDISCH_TAC
        `&1 / RR z' w0' lam = n / n' * &1 / RR z w0 lam + (n - n') / (n' * R)` THEN
      SIMP_TAC [real_div; REAL_MUL_LID] THEN DISCH_TAC THEN
      SIMP_TAC [REAL_INV_MUL];
      SIMP_TAC [real_div; REAL_INV_INV; REAL_INV_MUL] THEN REAL_ARITH_TAC]]);;




let BEAM_SI_REFLECTED = prove
 (`!R w0 w0' z z' n n' lam A B C D.
  
   &0 < n  /\ &0 < n' /\ ~(R = &0) /\ &0 < w0 /\ &0 < lam /\ &0 < w0' /\
  ~(z' = &0) /\ ~(z = &0) /\ ~(qq (z,w0,lam) = Cx (R / (&2))) /\
  (WW z' w0' lam = WW z w0 lam) /\
    (inv (RR z' w0' lam) = inv (RR z w0 lam) - &2 / R) /\
   cmat2x2 A B C D = interface_matrix n n' (spherical R) reflected  ==> 
   qq (z',w0',lam) =  (A * qq (z,w0,lam) + B)/ (C * qq (z,w0,lam) + D)`,
  SIMP_TAC [interface_matrix; CMAT2X2_EQ] THEN
  SIMP_TAC
    [COMPLEX_MUL_LID; COMPLEX_MUL_LZERO; COMPLEX_ADD_LID; complex_div;
     COMPLEX_MUL_RID; COMPLEX_INV_1; COMPLEX_MUL_LZERO; COMPLEX_MUL_RZERO;
     COMPLEX_ADD_RID] THEN
  REWRITE_TAC [GSYM complex_div] THEN REPEAT STRIP_TAC THEN
  ASSERT_TAC
    `qq (z',w0',lam) = qq (z,w0,lam) / (Cx (-- &2 / R) * qq (z,w0,lam) + Cx (&1)) <=> inv (qq (z',w0',lam)) = inv( qq (z,w0,lam) / (Cx (-- &2 / R) * qq (z,w0,lam) + Cx (&1)))` THENL
   [MATCH_MP_TAC (GSYM COMPLEX_INV_2) THEN
    ASM_SIMP_TAC [EQ_SYM_EQ; qq_QQ; Q_NZ] THEN
    ONCE_REWRITE_TAC [EQ_SYM_EQ; complex_div] THEN
    ONCE_REWRITE_TAC [complex_div] THEN
    SIMP_TAC [COMPLEX_ENTIRE; DE_MORGAN_THM] THEN
    ASM_SIMP_TAC [EQ_SYM_EQ; qq_QQ; Q_NZ] THEN ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
    IMP_REWRITE_TAC [COMPLEX_INV_NZ] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `~(Cx (-- &2 / R) * qq (z,w0,lam) + Cx (&1) = Cx (&0)) = ~(Cx (-- &2 / R) * qq (z,w0,lam) = --Cx (&1))`] THEN
    SIMP_TAC [CX_DIV; CX_NEG] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `~(--Cx (&2) / Cx R * qq (z,w0,lam) = --Cx (&1)) = ~(Cx (&2) / Cx R * qq (z,w0,lam) = Cx (&1))`] THEN
    IMP_REWRITE_TAC
      [COMPLEX_FIELD
         `~(Cx R = Cx(&0)) ==>
 ~(Cx (&2) / Cx R * qq (z,w0,lam) = Cx (&1)) = 
~( qq (z,w0,lam) = Cx R / Cx (&2))`;
       CX_INJ] THEN
    ASM_SIMP_TAC [GSYM CX_DIV; CX_INJ];
    ASM_REWRITE_TAC [] THEN
    SIMP_TAC
      [complex_div; COMPLEX_INV_MUL; COMPLEX_INV_INV; COMPLEX_ADD_LDISTRIB;
       COMPLEX_MUL_RID] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `inv (qq (z,w0,lam)) * Cx (-- &2 / R) * qq (z,w0,lam) = 
 (inv (qq (z,w0,lam))*qq (z,w0,lam))*Cx (-- &2 / R)`] THEN
    IMP_REWRITE_TAC [COMPLEX_MUL_LINV] THEN
    ASM_SIMP_TAC [Q_NZ; COMPLEX_MUL_LID; COMPLEX_ADD_SYM] THEN
    ONCE_REWRITE_TAC [COMPLEX_FIELD `!q. inv q  = Cx(&1)*inv q`] THEN
    REWRITE_TAC [GSYM complex_div] THEN ASM_SIMP_TAC [qq_QQ] THEN
    SIMP_TAC [QQ; complex_div; COMPLEX_INV_INV] THEN
    SIMP_TAC [COMPLEX_MUL_LID; COMPLEX_INV_MUL; COMPLEX_INV_INV] THEN
    SIMP_TAC [GSYM complex_div; COMPLEX_INV_1; COMPLEX_MUL_LID] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `Cx (-- &2 / R) +
 inv (Cx (RR z w0 lam)) - ii * Cx lam * inv (Cx pi) / Cx (WW z w0 lam pow 2) = 
 (inv (Cx (RR z w0 lam)) +  Cx (-- &2 / R))
 - ii * Cx lam * inv (Cx pi) / Cx (WW z w0 lam pow 2)`] THEN
    REWRITE_TAC [GSYM CX_INV; GSYM CX_ADD; GSYM CX_MUL; GSYM CX_DIV] THEN
    SIMP_TAC [COMPLEX_FIELD `!a b. a - ii * b = a + ii *(--b)`] THEN
    REWRITE_TAC [GSYM CX_NEG] THEN REWRITE_TAC [GSYM COMPLEX_TRAD] THEN
    REWRITE_TAC [COMPLEX_EQ] THEN REWRITE_TAC [GSYM CX_ADD; RE; IM] THEN
    SIMP_TAC [GSYM CX_INV; RE; IM] THEN
    ASM_SIMP_TAC
      [REAL_FIELD
         `~(lam = &0) /\ ~(inv pi = &0) ==> (--(lam * inv pi / WW z' w0' lam pow 2) =
 --(lam * inv pi / WW z w0 lam pow 2)) = (( inv ( WW z' w0' lam pow 2)) =
 ( inv (WW z w0 lam pow 2)))`;
       REAL_POS_NZ; REAL_INV_NZ; PI_POS] THEN
    ASM_SIMP_TAC [REAL_FIELD `! a (b:real). a + --b = a - b`] THEN
    REAL_ARITH_TAC]);;




let ABCD_LAW_OPTICAL_INTERFACE = prove
 (`!i ik z w0 z' w0' lam n n' A B C D. 
    is_valid_beam_at_inerface (z,w0) (z',w0') lam n n' i ik /\
    is_valid_interface i /\ &0 < lam /\
    cmat2x2 A B C D = interface_matrix n n' i ik ==>
    qq (z',w0',lam) =
    (A * qq (z,w0,lam) + B) / (C * qq (z,w0,lam) + D)`,
  MATCH_MP_TAC interface_IND THEN CONJ_TAC THENL
   [MATCH_MP_TAC interface_kind_IND THEN CONJ_TAC THEN
    REWRITE_TAC [is_valid_beam_at_inerface] THEN REPEAT STRIP_TAC THEN
    MATCH_MP_TAC ABCD_PLANE_INTEFACE THENL
     [EXISTS_TAC `transmitted`; EXISTS_TAC `reflected`] THEN
    EXISTS_TAC `n:real` THEN EXISTS_TAC `n':real` THEN
    ASM_SIMP_TAC [is_valid_beam_at_plane_inerf];
    GEN_TAC THEN MATCH_MP_TAC interface_kind_IND THEN CONJ_TAC THEN
    REWRITE_TAC
      [is_valid_beam_at_inerface; is_valid_interface; is_valid_single_beam] THEN
    REPEAT STRIP_TAC THENL
     [MATCH_MP_TAC BEAM_SI_TRANS; MATCH_MP_TAC BEAM_SI_REFLECTED] THEN
    EXISTS_TAC `a:real` THEN EXISTS_TAC `n:real` THEN EXISTS_TAC `n':real` THEN
    ASM_SIMP_TAC []]);;




 let INTERFACE_MATRIX_HEAD_INDEX = prove (
`!(n':real) (d':real)  i ik (a1:(optical_component)list). 
   interface_matrix n (head_index (a1,n',d')) i ik =
   interface_matrix n (head_index (a1,n',d')) i ik`,


REPEAT (MATCH_MP_TAC list_INDUCT ORELSE GEN_TAC) THEN
SIMP_TAC[head_index ;interface_matrix]);;


let CMAT_MUL_3 = prove(` !A1 B1 C1 D1 A2 B2 C2 D2 A3 B3 C3 D3.
   cmat2x2 A1 B1 C1 D1 **  cmat2x2 A2 B2 C2 D2 **  cmat2x2 A3 B3 C3 D3 =
   cmat2x2 (A1 * (A2 * A3 + B2 * C3) + B1 * (C2 * A3 + D2 * C3))
     (A1 * (A2 * B3 + B2 * D3) + B1 * (C2 * B3 + D2 * D3))
     (C1 * (A2 * A3 + B2 * C3) + D1 * (C2 * A3 + D2 * C3))
     (C1 * (A2 * B3 + B2 * D3) + D1 * (C2 * B3 + D2 * D3))`,

SIMP_TAC[CMAT2X2_MUL]);;



let Q_ABCD_ITERATE_2 = prove
 (`!A1 B1 C1 D1 A2 B2 C2 D2 q0 q1.  ~(C1 * q0 + D1 = Cx (&0))   /\
   cmat2x2 A B C D =  cmat2x2 A2 B2 C2 D2 ** cmat2x2 A1 B1 C1 D1    /\
   q1 = (A1*q0 + B1) / (C1*q0 + D1) ==>
   (A2*q1 + B2) / (C2*q1 + D2) = (A*q0 + B) / (C*q0 + D)`,
  SIMP_TAC [CMAT2X2_MUL; CMAT2X2_EQ] THEN REPEAT STRIP_TAC THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `!A2 B2 B1 C1 q0. (A2 * (A1 * q0 + B1) / (C1 * q0 + D1) + B2) = ( (A1*A2 * q0 +A2* B1) / (C1 * q0 + D1) + B2)`] THEN
  IMP_REWRITE_TAC [COMPLEX_FIELD `~(b = Cx(&0)) ==> a / b + c = (a + b*c)/b`] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `! a b t. ~(t = Cx(&0)) ==> (a/t)/(b/t) = 
                             (a/b)`] THEN
  BINOP_TAC THEN CONV_TAC COMPLEX_FIELD);;



let Q_ABCD_ITERATE_3 = prove
 (`!A1 B1 C1 D1 A2 B2 C2 D2 A3 B3 C3 D3 q0 q1 q2.  
  ~(C1 * q0 + D1 = Cx (&0))   /\
  ~(C2 * (A1 * q0 + B1) + D2 * (C1 * q0 + D1) = Cx (&0))
 /\
    cmat2x2 A B C D =  
    cmat2x2 A3 B3 C3 D3 ** cmat2x2 A2 B2 C2 D2 ** cmat2x2 A1 B1 C1 D1 /\
    q1 = (A1*q0 + B1) / (C1*q0 + D1) /\
    q2 = (A2*q1 + B2) / (C2*q1 + D2)
    ==>
    (A3*q2 + B3) / (C3*q2 + D3) = (A*q0 + B) / (C*q0 + D)`,
  SIMP_TAC [CMAT2X2_MUL; CMAT2X2_EQ] THEN REPEAT STRIP_TAC THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `! A1 A2 A3 B1 D1 C1 q0. 
A3 * (A2 * (A1 * q0 + B1) / (C1 * q0 + D1) + B2) = 
  A3 * ((A2 * A1 * q0 +  A2*B1) / (C1 * q0 + D1) + B2)`] THEN
  ASM_SIMP_TAC
    [COMPLEX_FIELD
       `~(C1 * q0 + D1 = Cx (&0)) ==> (A2 * (A1 * q0 + B1) / (C1 * q0 + D1) + B2) = ((A2 * (A1 * q0 + B1)) + B2*(C1 * q0 + D1))/(C1 * q0 + D1)`] THEN
  ASM_SIMP_TAC
    [COMPLEX_FIELD
       `~(C1 * q0 + D1 = Cx (&0)) ==> (C2 * (A1 * q0 + B1) / (C1 * q0 + D1) + B2) = ((C2 * (A1 * q0 + B1)) + B2*(C1 * q0 + D1))/(C1 * q0 + D1)`] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `! a b t. ~(t = Cx(&0)) ==> (a/t)/(b/t) = 
 

                            (a/b)`] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `~(b = Cx(&0)) ==> f*(a / b) + c = (f*a + b*c)/b`] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `! a b t. ~(t = Cx(&0)) ==> (a/t)/(b/t) = 
 

                            (a/b)`] THEN
  BINOP_TAC THEN CONV_TAC COMPLEX_FIELD);;




let ABCD_LAW_OPTICAL_SYSTEM = prove
 (`!sys beam aa bb cc dd. &0 < lam /\
    is_valid_optical_system sys /\ is_valid_beam_in_system beam lam sys /\
   cmat2x2 aa bb cc dd = system_composition sys  ==> 
      let ((z,w0),(z',w0'),rs) = beam in
      let (z_n,w0_n) = last_single_ray beam in

 qq (z_n,w0_n,lam) =
             (aa * qq (z,w0,lam) + bb) / (cc * qq (z,w0,lam) + dd)`,
  REWRITE_TAC [FORALL_OPTICAL_SYSTEM_THM; FORALL_BEAM_THM] THEN
  MATCH_MP_TAC list_INDUCT THEN CONJ_TAC THENL
   [REWRITE_TAC [FORALL_SINGLE_BEAM_THM; FORALL_OPTICAL_COMPONENT_THM] THEN
    REPEAT (MATCH_MP_TAC list_INDUCT ORELSE GEN_TAC) THEN CONJ_TAC THENL
     [ALL_TAC; SIMP_TAC [is_valid_beam_in_system]] THEN
    REWRITE_TAC
      [IS_VALID_OPTICAL_SYSTEM_REC; is_valid_beam_in_system; ALL;
       system_composition; head_index] THEN
    REPEAT LET_TAC THEN REPEAT (POP_ASSUM MP_TAC) THEN
    REWRITE_TAC [last_single_ray] THEN SIMP_TAC [PAIR_EQ] THEN
    ONCE_SIMP_TAC [EQ_SYM_EQ] THEN DISCH_TAC THEN ASM_SIMP_TAC [] THEN
    REWRITE_TAC [last_single_ray] THEN SIMP_TAC [PAIR_EQ] THEN
    REPEAT STRIP_TAC THEN ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
    MATCH_MP_TAC ABCD_FS THEN EXISTS_TAC `z:real` THEN EXISTS_TAC `w0:real` THEN
    ASM_SIMP_TAC [];
    ALL_TAC] THEN
  REWRITE_TAC [FORALL_OPTICAL_COMPONENT_THM] THEN
  ONCE_REWRITE_TAC [FORALL_FREE_SPACE_THM] THEN REPEAT GEN_TAC THEN DISCH_TAC THEN
  ONCE_REWRITE_TAC [FORALL_FREE_SPACE_THM] THEN
  REWRITE_TAC [FORALL_SINGLE_BEAM_THM] THEN REWRITE_TAC [system_composition] THEN
  REWRITE_TAC [is_valid_optical_system; is_valid_beam_in_system; ALL] THEN
  REPEAT (MATCH_MP_TAC list_INDUCT ORELSE GEN_TAC) THEN
  SIMP_TAC [is_valid_beam_in_system] THEN
  REWRITE_TAC [FORALL_SINGLE_BEAM_PAIR_THM] THEN
  REWRITE_TAC [FORALL_SINGLE_BEAM_THM] THEN
  REWRITE_TAC [is_valid_beam_in_system] THEN REPEAT STRIP_TAC THEN
  REPEAT LET_TAC THEN
  UNDISCH_TAC
    `(((z,w0),(z',w0'),CONS ((z'',w0''),z''',w0''') a1'):beam) =
      (z'''',w0''''),(z''''',w0'''''),rs` THEN
  SIMP_TAC [PAIR_EQ] THEN ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN REPEAT STRIP_TAC THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN ONCE_ASM_REWRITE_TAC [] THEN
  ASSERT_TAC
    `qq (z',w0',lam) =
(((free_space_matrix ((n:real),(d:real)))$1$1) * qq (z,w0,lam) + 
  ((free_space_matrix((n:real),(d:real)) )$1$2)) / 
(((free_space_matrix ((n:real),(d:real)))$2$1) * qq (z,w0,lam) + 
((free_space_matrix ((n:real),(d:real)))$2$2))` THENL
   [MATCH_MP_TAC ABCD_FS THEN EXISTS_TAC `n:real` THEN EXISTS_TAC `d:real` THEN
    ASM_SIMP_TAC [] THEN CCOMMON_TAC [free_space_matrix];
    ALL_TAC] THEN
  ASSERT_TAC
    `qq (z'',w0'',lam) =
(((interface_matrix n (head_index (a1,n',d')) i ik)$1$1) * qq (z',w0',lam) + 
((interface_matrix n (head_index (a1,n',d')) i ik)$1$2)) / 
(((interface_matrix n (head_index (a1,n',d')) i ik)$2$1) * qq (z',w0',lam) + 
((interface_matrix n (head_index 
((a1:((real#real)#interface#interface_kind)list),n',d')) i ik)$2$2))` THENL
   [MATCH_MP_TAC ABCD_LAW_OPTICAL_INTERFACE THEN EXISTS_TAC `i:interface` THEN
    EXISTS_TAC `ik:interface_kind` THEN EXISTS_TAC `n:real` THEN
    EXISTS_TAC
      `head_index((a1:((real#real)#interface#interface_kind)list),(n':real),(d':real))` THEN
    ASM_SIMP_TAC [] THEN
    UNDISCH_TAC `is_valid_optical_component (((n,d),i,ik):optical_component)` THEN
    SIMP_TAC
      [is_valid_optical_component; interface_matrix;
       INTERFACE_MATRIX_HEAD_INDEX] THEN
    REPEAT STRIP_TAC THEN SIMP_TAC [GSYM CMAT_2X2_EQ_COMPONENT];
    ALL_TAC] THEN
  POP_ASSUM MP_TAC THEN POP_ASSUM MP_TAC THEN REWRITE_TAC [EQ_SYM_EQ] THEN
  REPEAT STRIP_TAC THEN ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
  UNDISCH_TAC
    `!fs sr1 sr2 srs aa bb cc dd.
          &0 < lam /\
          is_valid_optical_system (a1,fs) /\
          is_valid_beam_in_system (sr1,sr2,srs) lam (a1,fs) /\
          cmat2x2 aa bb cc dd = system_composition (a1,fs)
          ==> (let (z,w0),(z',w0'),rs = sr1,sr2,srs in
               let z_n,w0_n = last_single_ray (sr1,sr2,srs) in
               qq (z_n,w0_n,lam) =
               (aa * qq (z,w0,lam) + bb) / (cc * qq (z,w0,lam) + dd))` THEN
  ONCE_REWRITE_TAC [FORALL_FREE_SPACE_THM] THEN
  REWRITE_TAC [is_valid_optical_system] THEN DISCH_TAC THEN
  POP_ASSUM (MP_TAC o SPEC `n':real`) THEN DISCH_TAC THEN
  POP_ASSUM (MP_TAC o SPEC `d':real`) THEN DISCH_TAC THEN
  POP_ASSUM (MP_TAC o SPEC `((z'':real),(w0'':real))`) THEN DISCH_TAC THEN
  POP_ASSUM (MP_TAC o SPEC `((z''':real),(w0''':real))`) THEN DISCH_TAC THEN
  POP_ASSUM (MP_TAC o SPEC `(a1':((real#real)#real#real)list)`) THEN
  DISCH_TAC THEN
  POP_ASSUM
    (MP_TAC o
     SPECL
       [`((system_composition (a1,n',d'))$1$1):complex`;
        `((system_composition (a1,n',d'))$1$2):complex`;
        `((system_composition (a1,n',d'))$2$1):complex`;
        `((system_composition (a1,n',d'))$2$2):complex`]) THEN
  ANTS_TAC THENL [ASM_SIMP_TAC [] THEN CCOMMON_TAC []; ALL_TAC] THEN
  REPEAT LET_TAC THEN
  UNDISCH_TAC
    `((z'',w0''),(z''',w0'''),a1'):beam =
      (z'''''',w0''''''),(z''''''',w0'''''''),rs'` THEN
  SIMP_TAC [PAIR_EQ] THEN ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN REPEAT STRIP_TAC THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN POP_ASSUM MP_TAC THEN
  ONCE_ASM_REWRITE_TAC [] THEN
  UNDISCH_TAC
    `last_single_ray ((z'''',w0''''),(z''''',w0'''''),rs) = z_n,w0_n` THEN
  ASM_REWRITE_TAC [] THEN REWRITE_TAC [LAST_SINGLE_RAY_REC] THEN
  UNDISCH_TAC
    `last_single_ray ((z'''''',w0''''''),(z''''''',w0'''''''),rs') =
      z_n',w0_n'` THEN
  ASM_REWRITE_TAC [] THEN SIMP_TAC [] THEN SIMP_TAC [PAIR_EQ] THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC [] THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN MATCH_MP_TAC Q_ABCD_ITERATE_3 THEN
  EXISTS_TAC `free_space_matrix ((n:real),(d:real))$1$1` THEN
  EXISTS_TAC `free_space_matrix ((n:real),(d:real))$1$2` THEN
  EXISTS_TAC `free_space_matrix ((n:real),(d:real))$2$1` THEN
  EXISTS_TAC `free_space_matrix ((n:real),(d:real))$2$2` THEN
  EXISTS_TAC
    `interface_matrix n (head_index 
((a1:((real#real)#interface#interface_kind)list),n',d')) i ik$1$1` THEN
  EXISTS_TAC
    `interface_matrix n (head_index 
((a1:((real#real)#interface#interface_kind)list),n',d')) i ik$1$2` THEN
  EXISTS_TAC
    `interface_matrix n (head_index 
((a1:((real#real)#interface#interface_kind)list),n',d')) i ik$2$1` THEN
  EXISTS_TAC
    `interface_matrix n (head_index 
((a1:((real#real)#interface#interface_kind)list),n',d')) i ik$2$2` THEN
  EXISTS_TAC `qq ((z':real),(w0':real),(lam:real)):complex` THEN CONJ_TAC THENL
   [SIMP_TAC [free_space_matrix; CMAT_2X2_COMPONENTS] THEN
    CONV_TAC COMPLEX_FIELD;
    ALL_TAC] THEN
  CONJ_TAC THENL
   [ALL_TAC;
    CONJ_TAC THENL
     [ASM_REWRITE_TAC [] THEN SIMP_TAC [GSYM CMAT_2X2_EQ_COMPONENT];
      REWRITE_TAC [EQ_SYM_EQ] THEN ASM_SIMP_TAC []]] THEN
  UNDISCH_TAC `is_valid_optical_component (((n,d),i,ik):optical_component)` THEN
  UNDISCH_TAC
    `is_valid_beam_at_inerface ((z':real),(w0':real))
      ((z'':real),(w0'':real))
      (lam:real)
      (n:real)
      (head_index
      ((a1:((real#real)#interface#interface_kind)list),(n':real),(d':real)))
      (i:interface)
      (ik:interface_kind)` THEN
  SPEC_TAC (`ik:interface_kind`,`ik:interface_kind`) THEN
  SPEC_TAC (`i:interface`,`i:interface`) THEN MATCH_MP_TAC interface_IND THEN
  CONJ_TAC THENL
   [MATCH_MP_TAC interface_kind_IND THEN
    SIMP_TAC [interface_matrix; CMAT_2X2_COMPONENTS; free_space_matrix] THEN
    SIMP_TAC
      [COMPLEX_MUL_LID; COMPLEX_MUL_LZERO; COMPLEX_MUL_RZERO;
       COMPLEX_ADD_LID; COMPLEX_ADD_RID; COMPLEX_MUL_RID; CX_DIV] THEN
    ASSERT_TAC `&0 <(head_index ((a1,n',d'):optical_system))` THENL
     [MATCH_MP_TAC IS_VALID_OPTICAL_SYSTEM_HEAD_INDEX THEN
      ASM_SIMP_TAC [is_valid_optical_system];
      ALL_TAC] THEN
    IMP_REWRITE_TAC
      [COMPLEX_FIELD `~(a = Cx(&0))/\ ~(b = Cx(&0)) ==> ~(a/b = Cx(&0))`] THEN
    SIMP_TAC [CX_INJ] THEN ASM_SIMP_TAC [REAL_POS_NZ] THEN
    SIMP_TAC [is_valid_optical_component; is_valid_free_space] THEN
    REAL_ARITH_TAC;
    ALL_TAC] THEN
  GEN_TAC THEN MATCH_MP_TAC interface_kind_IND THEN
  SIMP_TAC [interface_matrix; CMAT_2X2_COMPONENTS; free_space_matrix] THEN
  SIMP_TAC
    [COMPLEX_MUL_LID; COMPLEX_MUL_LZERO; COMPLEX_MUL_RZERO; COMPLEX_ADD_LID;
     COMPLEX_ADD_RID; COMPLEX_MUL_RID; CX_DIV] THEN
  REWRITE_TAC [is_valid_beam_at_inerface; is_valid_interface] THEN
  REWRITE_TAC [CX_NEG] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `--Cx (&2) / Cx a * (qq (z,w0,lam) + Cx d) + Cx (&1) =  Cx (&0) <=>
Cx (&2) / Cx a * (qq (z,w0,lam) + Cx d) =  Cx (&1)`] THEN
  REWRITE_TAC
    [is_valid_optical_component; is_valid_free_space; is_valid_interface] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `!a q. ~(Cx(&0) = Cx a) ==>
 ~(Cx (&2) / Cx a * q = Cx (&1)) = 
~( q = Cx a / Cx (&2))`;
     CX_INJ] THEN
  UNDISCH_TAC
    `is_valid_beam_in_free_space ((z:real),(w0:real)) 
               ((z':real),(w0':real)) ((n:real),(d:real))` THEN
  SIMP_TAC [is_valid_beam_in_free_space] THEN DISCH_TAC THEN CONJ_TAC THENL
   [ALL_TAC;
    SIMP_TAC [qq; CX_ADD] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `(Cx z + Cx d) + ii * Cx (rayleigh_range w0 lam) = (Cx z + ii * Cx (rayleigh_range w0 lam)) + Cx d`;
       CX_DIV] THEN
    SIMP_TAC [CX_INJ]] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(Cx ((n - head_index (a1,n',d')) / head_index (a1,n',d') * &1 / a) *
       (qq (z,w0,lam) + Cx d) +
       Cx n / Cx (head_index (a1,n',d')) =
       Cx (&0)) = (Cx ((n - head_index (a1,n',d')) / head_index (a1,n',d') * &1 / a) *
       (qq (z,w0,lam) + Cx d) = --
       Cx n / Cx (head_index (a1,n',d')))`] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `~(b = Cx(&0)) ==> (a  =
   Cx (&1) / b) = (a * b  =
   Cx (&1))`] THEN
  IMP_REWRITE_TAC [Q_NZ; is_valid_single_beam] THEN REWRITE_TAC [CX_DIV] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `~(b = Cx(&0))

 ==> ((a / b)*c  =
   Cx (&1))  = (a*c = (b))`] THEN
  SIMP_TAC [MESON [] `a /\ b ==> c ==> d <=> a /\ b /\c ==> d`] THEN
  SIMP_TAC [CX_INJ; REAL_ENTIRE] THEN
  IMP_REWRITE_TAC [REAL_FIELD `&0 < n  /\ ~(&0 = a) ==> ~(n = &0 \/ a = &0)`] THEN
  IMP_REWRITE_TAC [REAL_FIELD `k / b* (&1/a) = (k/ (b*a))`] THEN
  REWRITE_TAC [CX_DIV] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `~(b = Cx(&0)) 

 ==> ((a / b)*c  = t) =
      (a*c = t*b)`] THEN
  SIMP_TAC [CX_INJ; REAL_POS_NZ; REAL_ENTIRE] THEN REWRITE_TAC [CX_MUL] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `~(b = Cx(&0)) ==> (--a / b) * b*d   = --a*d`] THEN
  SIMP_TAC [CX_INJ; REAL_POS_NZ; REAL_ENTIRE] THEN
  SIMP_TAC [COMPLEX_FIELD `a*b = --c*d <=> --a*b = c*d`] THEN
  SIMP_TAC [GSYM CX_NEG] THEN SIMP_TAC [qq; CX_ADD] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(Cx z + Cx d) + ii * Cx (rayleigh_range w0 lam) = (Cx z + ii * Cx (rayleigh_range w0 lam)) + Cx d`]);;
